/*
 *  dvipr
 *      TeX Printer driver for The Canon LIPS III/IVc, CaPSL III printer
 *
 *      Developed at:
 *          Computer Systems Research Group
 *          Department of Information Science
 *          Fukui University
 *          9-1, Bunkyo-3, Fukui, 910 Japan
 *
 *      Authors:
 *          Yasumasa Koizumi, Atsushi Takagi, Kenichi Sugimoto,
 *          and Minato Kawaguti
 *
 *  The program may be copied and redistributed freely in the spirit of
 *  Free Software Foundation's "Copy Left", provided this title page is
 *  included in all of the source files involved.
 *
 *  Reports on porting to new platforms, bug fixes, etc., are highly
 *  appreciated. e-mail of this kind of communication shhould be addressedd
 *  to:
 *      dvipr@i1nws1.fukui-u.ac.jp
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "newtypes.h"
#include "gvars.h"
#include "pdl.h"
#include "color.h"
#include "error.h"
#include "funcdef.h"

#define	TOKEN_STR	" \t,"

static Color	*pColorRoot, *pColorStack;
static Color	CurrentColor;

static Color *CreateColor(Color** pRoot);
static Color *CreateStackColor(void);
static BOOL PushColor(char *ptr);
static BOOL PopColor(void);
static BOOL SearchColor(char *pszName, Color *pColor);
static BOOL ReadColorValue(char *ptr, Color* pColor);

BOOL Specialcolor(char *pszSpecial)
{
	char	*ptr;

	ptr = strtok(pszSpecial, TOKEN_STR);
	if ((ptr = strtok(NULL, TOKEN_STR)) == NULL)
		return FALSE;
	if (!strcmp(ptr, "push"))
		return PushColor(&ptr[5]);
	else if (!strcmp(ptr, "pop"))
		return PopColor();
	return FALSE;
}

BOOL SetupColor(void)
{
	int		i;
	char	aszColorTableFileName[FILENAME_MAX], acBuff[BUFSIZ];
	char	*pszName, *ptr;
	Color	*pNewColor, color;
	FILE	*fp;

	pColorRoot = pColorStack = NULL;
	sprintf(aszColorTableFileName, "%s/dvipdl.psc", DVIPR_DATA_DIR);
	if ((fp = fopen(aszColorTableFileName, "rt")) == NULL) {
		Error2(ERR_COLOR_NOTABLEFILE, aszColorTableFileName);
		return FALSE;
	}
	while (fgets(acBuff, BUFSIZ, fp) != NULL) {
		*strchr(acBuff, '\n') = '\0';
		if (acBuff[0] == '#')
			continue;
		if ((pszName = strtok(acBuff, TOKEN_STR)) == NULL)
			continue;
		if ((ptr = strtok(NULL, "")) == NULL) {
			Error2(ERR_COLOR_BADFORMAT, aszColorTableFileName);
			continue;
		}
		if (!ReadColorValue(ptr, &color)) {
			Error2(ERR_COLOR_BADFORMAT, aszColorTableFileName);
			continue;
		}
		if ((pNewColor = CreateColor(&pColorRoot)) == NULL)
			return FALSE;
		strcpy(pNewColor->acName, pszName);
		pNewColor->Type = color.Type;
		for (i = 0; i < 4; i++)
			pNewColor->dwValue[i] = color.dwValue[i];
	}
	fclose(fp);
	return TRUE;
}

static Color *CreateColor(Color** pRoot)
{
	Color* pNewColor, *ptr;

	if ((pNewColor = (Color*)malloc(sizeof (Color))) == NULL) {
		Error(ERR_NOMEMORY);
		return NULL;
	}
	pNewColor->next = NULL;
	if (*pRoot == NULL)
		*pRoot = pNewColor;
	else {
		for (ptr = *pRoot; ptr->next; ptr = ptr->next);
		ptr->next = pNewColor;
	}
	return pNewColor;
}

static Color *CreateStackColor(void)
{
	Color* pNewColor;

	if ((pNewColor = (Color*)malloc(sizeof (Color))) == NULL) {
		Error(ERR_NOMEMORY);
		return NULL;
	}
	pNewColor->next = pColorStack;
	pColorStack = pNewColor;
	return pNewColor;
}

static BOOL PushColor(char *ptr)
{
	int	i;
	Color	*pNewColor;

	ptr = strtok(NULL, "");
	if ((pNewColor = CreateStackColor()) == NULL)
		return FALSE;
	pNewColor->Type = CurrentColor.Type;
	for (i = 0; i < 4; i++)
		pNewColor->dwValue[i] = CurrentColor.dwValue[i];
	if (!ReadColorValue(ptr, &CurrentColor)) {
		Error2(ERR_COLOR_UNDEFINEDCOLOR, ptr);
		return FALSE;
	}
	gpfnPdlSetColor(&CurrentColor);
	return TRUE;
}

static BOOL PopColor(void)
{
	if (pColorStack == NULL) {	
		Error(ERR_COLOR_NOSTACK);
		return FALSE;
	}
	memcpy(&CurrentColor, pColorStack, sizeof (Color));
	pColorStack = pColorStack->next;
	gpfnPdlSetColor(&CurrentColor);
	return TRUE;
}

static BOOL SearchColor(char *pszName, Color *pColor)
{
	Color	*ptr;

	for (ptr = pColorRoot; ptr; ptr = ptr->next)
		if (!strcmp(ptr->acName, pszName)) {
			memcpy(pColor, ptr, sizeof (Color));
			return TRUE;
		}
	return FALSE;
}

static BOOL ReadColorValue(char *ptr, Color* pColor)
{
	int	i;
	char	*pszType, *pszValue[4];
	double	lfValue;

	if ((pszType = strtok(ptr, TOKEN_STR)) == NULL)
		return FALSE;
	if (SearchColor(pszType, pColor))
		return TRUE;
	if (!strcmp(pszType, "rgb"))
		pColor->Type = COLOR_RGB;
	else if (!strcmp(pszType, "cmyk"))
		pColor->Type = COLOR_CMYK;
	else if (!strcmp(pszType, "gray"))
		pColor->Type = COLOR_GRAY;
	else {
		Error(ERR_COLOR_UNKNOWN_CLASS);
		return FALSE;
	}
	for (i = 0; i < (pColor->Type == COLOR_CMYK ? 4 : (pColor->Type == COLOR_RGB ? 3 : 1)); i++) {
		if ((pszValue[i] = strtok(NULL, TOKEN_STR)) == NULL)
			return FALSE;
		lfValue = atof(pszValue[i]);
		if (!(0. <= lfValue && lfValue <= 1.)) {
			Error(ERR_COLOR_BADPARAMETER);
			return FALSE;
		}
		pColor->dwValue[i] = (DWORD)(atof(pszValue[i]) * (double)COLOR_SCALED);
	}
	return TRUE;
}
