/*
 *  dvipr
 *      TeX Printer driver for The Canon LIPS III/IVc, CaPSL III printer
 *
 *      Developed at:
 *          Computer Systems Research Group
 *          Department of Information Science
 *          Fukui University
 *          9-1, Bunkyo-3, Fukui, 910 Japan
 *
 *      Authors:
 *          Yasumasa Koizumi, Atsushi Takagi, Kenichi Sugimoto,
 *          and Minato Kawaguti
 *
 *  The program may be copied and redistributed freely in the spirit of
 *  Free Software Foundation's "Copy Left", provided this title page is
 *  included in all of the source files involved.
 *
 *  Reports on porting to new platforms, bug fixes, etc., are highly
 *  appreciated. e-mail of this kind of communication shhould be addressedd
 *  to:
 *      dvipr@i1nws1.fukui-u.ac.jp
 */

#include <stdio.h>
#include <memory.h>
#include <stdlib.h>
#include <unistd.h>
#include "newtypes.h"
#include "gvars.h"
#include "file.h"
#include "dvi.h"
#include "dvipdl.h"
#include "pdl.h"
#include "special.h"
#include "font.h"
#include "font_jap.h"
#include "funcdef.h"

static DviRegister*	pDviReg;
static PageTable			*pPageTableRoot;
static unsigned WORD		wTotalPages;
static unsigned DWORD		dwNum, dwDen;
static DWORD		dwFirstPagePos, dwLastPagePos;
static DWORD dwMagnification;

static BOOL ScanPage(FILE *fp, WORD wPageNum);
static BOOL PrintPage(FILE *fp, WORD wPageNum);
static BOOL ReadDviPostamble(FILE *fp);
static BOOL MakePageTable(FILE *fp);
static DWORD GetPageTable(WORD wPageNum);
static BOOL ReadDviBOP(FILE* fp);
static BOOL ReadDviFNT_DEF(FILE* fp);
static BOOL ReadDviPOST(FILE *fp);
static BOOL ReadDviPOST_POST(FILE *fp, DWORD* loc);
static BOOL ReadDviPreamble(FILE *fp);

BOOL ReadDvi(char* pszDviFileName)
{
	unsigned WORD	i, nEndPage;
	FILE*			fpDvi;

	if (!strcmp(pszDviFileName, "-"))
		fpDvi = stdin;
	else
		if ((fpDvi = fopen(pszDviFileName, "rb")) == NULL)
			if ((fpDvi = fopen(pszDviFileName, "rb")) == NULL) {
				Error(ERR_DVI_NOTFOUND);
				return FALSE;
			}
	if (!ReadDviPreamble(fpDvi))
		return FALSE;
	if (!gpfnPdlPrintStart())
		return FALSE;
	if (!ReadDviPostamble(fpDvi))
		return FALSE;
	ResetCounter();
	nEndPage = wTotalPages > gnEndPage ? gnEndPage : wTotalPages;
	for (i = gnStartPage; i <= nEndPage; i++) {
		if (gfPrintEvenPage && (i % 2))
			continue;
		if (gfPrintOddPage && !(i % 2))
			continue;
		ScanPage(fpDvi, i);
	}
	gpfnPdlPrintInfo();
	for (i = gnStartPage; i <= nEndPage; i++) {
		if (gfPrintEvenPage && (i % 2))
			continue;
		if (gfPrintOddPage && !(i % 2))
			continue;
		if (!PrintPage(fpDvi, i))
			return FALSE;
		gnOutPapers++;
	}
	return gpfnPdlPrintEnd();
}

static BOOL ScanPage(FILE *fp, WORD wPageNum)
{
	unsigned	cmd;
	BOOL		bDirection;
	Fonts		*pFont = NULL;

	bDirection = FALSE;
	FpSeek(fp, GetPageTable(wPageNum), SEEK_SET);
	FpSeek(fp, 1 + 4 * 10 + 4, SEEK_CUR);
	while (TRUE) {
		cmd = FpRead1(fp);
		if (DVI_FNT_NUM_0 <= cmd && cmd <= DVI_FNT_NUM_63) {
			pFont = GetFonts(cmd - DVI_FNT_NUM_0);
			continue;
		} else if (DVI_FNT1 <= cmd && cmd <= DVI_FNT4) {
			pFont = GetFonts(FpReadN(fp, cmd - DVI_FNT1 + 1));
			continue;
		} else if (DVI_FNT_DEF1 <= cmd && cmd <= DVI_FNT_DEF4) {
			FpReadN(fp, cmd - DVI_FNT_DEF1 + 1);
			FpSeek(fp, 4 + 4 + 4, SEEK_CUR);
			FpSeek(fp, FpRead1(fp) + FpRead1(fp), SEEK_CUR);
			continue;
		} else if (cmd <= DVI_SET_CHAR_127) {
		/* I coded following first, but compiler warn it that always >= 0 */
		/* if (DVI_SET_CHAR_0 <= cmd && cmd <= DVI_SET_CHAR_127) { */
/*			if (!gfUseInternalFont)*/
			if (!(pFont->Type == FONT_INTERNAL_1BYTE || pFont->Type == FONT_INTERNAL || pFont->Type == FONT_DNPINTERNAL))
				GetGryph(pFont, cmd - DVI_SET_CHAR_0)->dwCounter++;
			continue;
		} else if (DVI_SET1 <= cmd && cmd <= DVI_SET4) {
			SelectJapaneseFont(FpReadN(fp, cmd - DVI_SET1 + 1), pFont);
			continue;
		} else if (DVI_RIGHT1 <= cmd && cmd <= DVI_RIGHT4) {
			FpSeek(fp, cmd - DVI_RIGHT1 + 1, SEEK_CUR);
			continue;
		} else if (DVI_DOWN1 <= cmd && cmd <= DVI_DOWN4) {
			FpSeek(fp, cmd - DVI_DOWN1 + 1, SEEK_CUR);
			continue;
		} else if (DVI_W1 <= cmd && cmd <= DVI_W4) {
			FpSeek(fp, cmd - DVI_W1 + 1, SEEK_CUR);
			continue;
		} else if (DVI_X1 <= cmd && cmd <= DVI_X4) {
			FpSeek(fp, cmd - DVI_X1 + 1, SEEK_CUR);
			continue;
		} else if (DVI_Y1 <= cmd && cmd <= DVI_Y4) {
			FpSeek(fp, cmd - DVI_Y1 + 1, SEEK_CUR);
			continue;
		} else if (DVI_Z1 <= cmd && cmd <= DVI_Z4) {
			FpSeek(fp, cmd - DVI_Z1 + 1, SEEK_CUR);
			continue;
		} else if (DVI_XXX1 <= cmd && cmd <= DVI_XXX4) {
			FpSeek(fp, FpReadN(fp, cmd - DVI_XXX1 + 1), SEEK_CUR);
			continue;
		} else if (cmd == DVI_EOP)
			break;
		else 
			switch (cmd) {
				case DVI_SET_RULE:
				case DVI_PUT_RULE:
					FpSeek(fp, 8, SEEK_CUR);
					break;
				case DVI_W0:
				case DVI_X0:
				case DVI_Y0:
				case DVI_Z0:
				case DVI_PUSH:
				case DVI_POP:
				case DVI_NOP:
					break;
				case DVI_DIR:
					bDirection = FpRead1(fp) ? TRUE : FALSE;
					break;
				default:
fprintf(stderr, "unknown cmd = %x\n", cmd);
			}
	}
	return TRUE;
}

static BOOL PrintPage(FILE *fp, WORD wPageNum)
{
	DWORD		a, b;
	BOOL		bDirection;
	unsigned	cmd;

	gpfnPdlNewPage(wPageNum);
	FpSeek(fp, GetPageTable(wPageNum), SEEK_SET);
	ReadDviBOP(fp);
	while (TRUE) {
		cmd = FpRead1(fp);
		if (DVI_FNT_NUM_0 <= cmd && cmd <= DVI_FNT_NUM_63) {
			xDviReg.f = cmd - DVI_FNT_NUM_0;
			gpfnPdlChangeFont(xDviReg.f);
			continue;
		} else if (DVI_FNT1 <= cmd && cmd <= DVI_FNT4) {
			xDviReg.f = FpReadN(fp, cmd - DVI_FNT1 + 1);
			gpfnPdlChangeFont(xDviReg.f);
			continue;
		} else if (DVI_FNT_DEF1 <= cmd && cmd <= DVI_FNT_DEF4) {
			FpReadN(fp, cmd - DVI_FNT_DEF1 + 1);
			FpSeek(fp, 4 + 4 + 4, SEEK_CUR);
			FpSeek(fp, FpRead1(fp) + FpRead1(fp), SEEK_CUR);
			continue;
		} else if (cmd <= DVI_SET_CHAR_127) {
		/* I coded following first, but compiler warn that it is always >= 0 */
		/* if (DVI_SET_CHAR_0 <= cmd && cmd <= DVI_SET_CHAR_127) { */
			if (!xDviReg.d)
				xDviReg.h += PutChar(cmd - DVI_SET_CHAR_0, xDviReg.f);
			else
				xDviReg.v += PutChar(cmd - DVI_SET_CHAR_0, xDviReg.f);
			continue;
		} else if (DVI_SET1 <= cmd && cmd <= DVI_SET4) {
			if (!xDviReg.d)
				xDviReg.h += PutChar(FpReadN(fp, cmd - DVI_SET1 + 1), xDviReg.f);
			else
				xDviReg.v += PutChar(FpReadN(fp, cmd - DVI_SET1 + 1), xDviReg.f);
			continue;
		} else if (DVI_RIGHT1 <= cmd && cmd <= DVI_RIGHT4) {
			if (!xDviReg.d)
				xDviReg.h += FpReadN(fp, cmd - DVI_RIGHT1 + 1);
			else
				xDviReg.v += FpReadN(fp, cmd - DVI_RIGHT1 + 1);
			continue;
		} else if (DVI_DOWN1 <= cmd && cmd <= DVI_DOWN4) {
			if (!xDviReg.d)
				xDviReg.v += FpReadN(fp, cmd - DVI_DOWN1 + 1);
			else
				xDviReg.h -= FpReadN(fp, cmd - DVI_DOWN1 + 1);
			continue;
		} else if (DVI_W1 <= cmd && cmd <= DVI_W4) {
			xDviReg.w = FpReadN(fp, cmd - DVI_W1 + 1);
			if (!xDviReg.d)
				xDviReg.h += xDviReg.w;
			else
				xDviReg.v += xDviReg.w;
			continue;
		} else if (DVI_X1 <= cmd && cmd <= DVI_X4) {
			xDviReg.x = FpReadN(fp, cmd - DVI_X1 + 1);
			if (!xDviReg.d)
				xDviReg.h += xDviReg.x;
			else
				xDviReg.v += xDviReg.x;
			continue;
		} else if (DVI_Y1 <= cmd && cmd <= DVI_Y4) {
			xDviReg.y = FpReadN(fp, cmd - DVI_Y1 + 1);
			if (!xDviReg.d)
				xDviReg.v += xDviReg.y;
			else
				xDviReg.h -= xDviReg.y;
			continue;
		} else if (DVI_Z1 <= cmd && cmd <= DVI_Z4) {
			xDviReg.z = FpReadN(fp, cmd - DVI_Z1 + 1);
			if (!xDviReg.d)
				xDviReg.v += xDviReg.z;
			else
				xDviReg.h -= xDviReg.z;
			continue;
		} else if (DVI_XXX1 <= cmd && cmd <= DVI_XXX4) {
			DoSpecial(fp, FpReadN(fp, cmd - DVI_XXX1 + 1));
			continue;
		} else if (cmd == DVI_EOP)
			break;
		else 
			switch (cmd) {
				case DVI_SET_RULE:
					a = FpRead4(fp);
					b = FpRead4(fp);
					gpfnPdlPutRule(a, b);
					if (!xDviReg.d)
						xDviReg.h += b;
					else
						xDviReg.v += b;
					break;
				case DVI_PUT_RULE:
					a = FpRead4(fp);
					b = FpRead4(fp);
					gpfnPdlPutRule(a, b);
					break;
				case DVI_W0:
					if (!xDviReg.d)
						xDviReg.h += xDviReg.w;
					else
						xDviReg.v += xDviReg.w;
					break;
				case DVI_X0:
					if (!xDviReg.d)
						xDviReg.h += xDviReg.x;
					else
						xDviReg.v += xDviReg.x;
					break;
				case DVI_Y0:
					if (!xDviReg.d)
						xDviReg.v += xDviReg.y;
					else
						xDviReg.h -= xDviReg.y;
					break;
				case DVI_Z0:
					if (!xDviReg.d)
						xDviReg.v += xDviReg.z;
					else
						xDviReg.h -= xDviReg.z;
					break;
				case DVI_PUSH:
					PushRegister();
					break;
				case DVI_POP:
					bDirection = xDviReg.d;
					PopRegister();
					if (bDirection != xDviReg.d)
						gpfnPdlChangeFont(xDviReg.f);
					break;
				case DVI_DIR:
					xDviReg.d = FpRead1(fp) ? TRUE : FALSE;
					if (xDviReg.f)
						gpfnPdlChangeFont(xDviReg.f);
					break;
				case DVI_NOP:
					continue;
				default:
fprintf(stderr, "unknown cmd = %x\n", cmd);
			}
	}
	gpfnPdlFeedPage();
	return TRUE;
}

BOOL PushRegister(void)
{
	DviRegister* reg;

	if ((reg = (DviRegister*)malloc(sizeof (DviRegister))) == NULL) {
		Error(ERR_NOMEMORY);
		return FALSE;
	}
	reg->h = xDviReg.h;	reg->v = xDviReg.v;
	reg->w = xDviReg.w;
	reg->x = xDviReg.x;	reg->y = xDviReg.y;
	reg->z = xDviReg.z;
	reg->d = xDviReg.d;
	reg->next = pDviReg;
	pDviReg = reg;
	return TRUE;
}

BOOL PopRegister(void)
{
	if (pDviReg == NULL) {
		Error(ERR_DVI_NOSTACK);
		return FALSE;
	}
	xDviReg.h = pDviReg->h; xDviReg.v = pDviReg->v;
	xDviReg.w = pDviReg->w;
	xDviReg.x = pDviReg->x; xDviReg.y = pDviReg->y;
	xDviReg.z = pDviReg->z;
	xDviReg.d = pDviReg->d;
	xDviReg.next = pDviReg->next;
	free(pDviReg);
	pDviReg = xDviReg.next;
	return TRUE;
}

static BOOL ReadDviPostamble(FILE *fp)
{
	DWORD	dwPostLoc;

	ReadDviPOST_POST(fp, &dwPostLoc);
	FpSeek(fp, dwPostLoc, SEEK_SET);
	ReadDviPOST(fp);
	if (!ReadDviFNT_DEF(fp))
		return FALSE;
	return MakePageTable(fp);
}

static BOOL MakePageTable(FILE *fp)
{
	DWORD		pos;
	PageTable	*pNewPageTable, *prev;

	prev = NULL;
	pos = dwLastPagePos;
	do {
		FpSeek(fp, pos, SEEK_SET);
		if (FpRead1(fp) != DVI_BOP) {
			Error(ERR_DVI_INVALID_FORMAT);
			return FALSE;
		}
		if ((pNewPageTable = (PageTable*)malloc(sizeof (PageTable))) == NULL) { 
			Error(ERR_NOMEMORY);
			return FALSE;
		}
		pNewPageTable->dwPos = pos;
		pNewPageTable->next = prev;
		prev = pNewPageTable;
		FpSeek(fp, 4 * 10, SEEK_CUR);
		pos = FpRead4(fp);
	} while (pos != -1);
	pPageTableRoot = prev;
	return TRUE;
}

static DWORD GetPageTable(WORD wPageNum)
{
	PageTable	*ptr;

	for (ptr = pPageTableRoot, --wPageNum; ptr && wPageNum; ptr = ptr->next, wPageNum--);
	return ptr->dwPos;
}

static BOOL ReadDviBOP(FILE* fp)
{
	int		i;
	WORD	wCount[10];

	if (FpRead1(fp) != DVI_BOP) {
		Error(ERR_DVI_INVALID_FORMAT);
		return FALSE;
	}
	for (i = 0; i < 10; i++)
		wCount[i] = FpRead4(fp);
	FpRead4(fp);
	xDviReg.h = 0;
	xDviReg.v = 0;
	xDviReg.w = 0;
	xDviReg.x = 0;
	xDviReg.y = 0;
	xDviReg.z = 0;
	xDviReg.f = 0;
	xDviReg.d = FALSE;
	pDviReg = NULL;
	return TRUE;
}

static BOOL ReadDviFNT_DEF(FILE* fp)
{
	int		cmd, i;
	char*	ptr;
	Fonts	*font, *font2;

	while (TRUE) {
		cmd = FpRead1(fp);
		if (cmd == DVI_POST_POST)
			break;
		if (!(DVI_FNT_DEF1 <= cmd && cmd <= DVI_FNT_DEF4)) {
			Error(ERR_DVI_INVALID_FORMAT);
			return FALSE;
		}
		font = CreateFont();
		font->dwFontNum = FpReadN(fp, cmd - DVI_FNT_DEF1 + 1);
		font->bDirection = FALSE;
		font->dwCheckSum = FpRead4(fp);
		font->dwScaled = FpRead4(fp);
		font->dwDesignSize = FpRead4(fp);
		font->wDPI =(WORD)((double)gwDPI * ((double)gdwMagnification / 1000.) * (double)((int)(((double)font->dwScaled / (double)font->dwDesignSize * 1000.) + .5) / 1000.) + .5);
		memset(font->acDir, 0, FILENAME_MAX);
		memset(font->acName, 0, FONTNAME_MAX);
		for (ptr = font->acDir, i = FpRead1(fp); i; i--)
			*ptr++ = FpRead1(fp);
		for (ptr = font->acName, i = FpRead1(fp); i; i--)
			*ptr++ = FpRead1(fp);
		if (gfUseInternalFont)
			if (ReplaceJapaneseFont(font))
				continue;
		if (!SelectFont(font))
			return FALSE;
		if (gnDviVersion == 3) {
			font2 = CreateFont();
			CopyFont(font2, font);
			font2->bDirection = TRUE;
			if (!SelectFont(font2))
				return FALSE;
			RotateFont(font2);
			font->dwFontNum += 1024;
		}
	}
	return TRUE;
}

static BOOL ReadDviPOST(FILE* fp)
{
	if (FpRead1(fp) != DVI_POST) {
		Error(ERR_DVI_INVALID_FORMAT);
		return FALSE;
	}
	dwLastPagePos = FpRead4(fp);
	if (dwNum != FpRead4(fp)) {
		Error(ERR_DVI_INVALID_FORMAT);
		return FALSE;
	}
	if (dwDen != FpRead4(fp)) {
		Error(ERR_DVI_INVALID_FORMAT);
		return FALSE;
	}
	if (dwMagnification != FpRead4(fp)) {
		Error(ERR_DVI_INVALID_FORMAT);
		return FALSE;
	}
	FpRead4(fp);
	FpRead4(fp);
	FpRead2(fp);
	wTotalPages = FpRead2(fp);
	return TRUE;
}

static BOOL ReadDviPOST_POST(FILE *fp, DWORD* loc)
{
	FpSeek(fp, -1, SEEK_END);
	while ((gnDviVersion = FpRead1(fp)) == DVI_TAIL)
		FpSeek(fp, -2, SEEK_CUR);
	if (gnDviVersion > DVI_VERSION) {
		Error(ERR_DVI_UNKNOWN_VERSION);
		return FALSE;
	}
	FpSeek(fp, -5, SEEK_CUR);
	if (loc != NULL)
		*loc = FpRead4(fp);
	return TRUE;
}

static BOOL ReadDviPreamble(FILE *fp)
{
	if (FpRead1(fp) != DVI_PRE) {
		Error(ERR_DVI_INVALID_FORMAT);
		return FALSE;
	}
	if (FpRead1(fp) > DVI_VERSION) {
		Error(ERR_DVI_UNKNOWN_VERSION);
		return FALSE;
	}
	dwNum = FpRead4(fp);
	dwDen = FpRead4(fp);
	dwMagnification = FpRead4(fp);
	gdwMagnification = (DWORD)((double)gdwMagnification * (double)dwMagnification / 1000.);
	/* The unit of dwNum / dwDen * 10^-7 is [m]. */
	/* So, I convert it to [inch].               */
	glfDimension = (double)dwNum / (double)dwDen * Pow(10., -7) * IPM;
	FpSeek(fp, FpRead1(fp), SEEK_CUR);
	dwFirstPagePos = FpTell(fp);
	ReadDviPOST_POST(fp, NULL);
	return TRUE;
}

double Pow(double a, int n)
{
	if (n < 0)
		return 1. / a * Pow(a, n + 1);
	else if (n > 0)
		return a * Pow(a, n - 1);
	return 1.;
}

DWORD PowDW(DWORD a, unsigned DWORD b)
{
	if (b)
		return a * PowDW(a, b - 1);
	return 1;
}
