/*
 *  dvipr
 *      TeX Printer driver for The Canon LIPS III/IVc, CaPSL III printer
 *
 *      Developed at:
 *          Computer Systems Research Group
 *          Department of Information Science
 *          Fukui University
 *          9-1, Bunkyo-3, Fukui, 910 Japan
 *
 *      Authors:
 *          Yasumasa Koizumi, Atsushi Takagi, Kenichi Sugimoto,
 *          and Minato Kawaguti
 *
 *  The program may be copied and redistributed freely in the spirit of
 *  Free Software Foundation's "Copy Left", provided this title page is
 *  included in all of the source files involved.
 *
 *  Reports on porting to new platforms, bug fixes, etc., are highly
 *  appreciated. e-mail of this kind of communication shhould be addressedd
 *  to:
 *      dvipr@i1nws1.fukui-u.ac.jp
 */

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "newtypes.h"
#include "dvi.h"
#include "pdl.h"
#include "dvipdl.h"
#include "color.h"
#include "gvars.h"
#include "error.h"
#include "funcdef.h"
#ifdef DVIPR_USE_KPATHSEA
#include <kpathsea/progname.h>
#include <kpathsea/proginit.h>
#include <kpathsea/tex-file.h>
#endif

char	*gpszWorkDirectory;
int		gnOutPapers;
int		gnPaperFeedMode;
BOOL	gfUseInternalFont;
BOOL	gfEnableFontDownLoad;
BOOL	gfForceFontDownLoad;
int		gnThresholdCharacters;
int		gnThresholdMemory;

static char gaszDviFileName[FILENAME_MAX + 1];
static BOOL gfTitleDisp;
static char *pszAccountFile, *pszRemoteUser, *pszRemoteHost;

static BOOL AnalizeArgument(int argc, char *argv[]);
/* static BOOL ReadEnvFile(void); */
static char *GetNextOption(int argc, char *argv[], int ct);
static BOOL Initialize(char *argv[]);
static BOOL Setup(char *argv[]);
static void PrintTitle(void);
static void PrintUsage(void);
static BOOL WriteAccount(void);

int main(int argc, char *argv[])
{
	if (!Initialize(argv))
		return -1;
/*	ReadEnvFile();*/
	if (!AnalizeArgument(argc, argv)) {
		PrintUsage();
		return -2;
	}
	if (gfTitleDisp)
		PrintTitle();
	Setup(argv);
	ReadDvi(gaszDviFileName);
	WriteAccount();
	return 0;
}

static BOOL Initialize(char *argv[])
{
	gpszWorkDirectory = NULL;
	gfUseInternalFont = TRUE;
	gfEnableFontDownLoad = TRUE;
	gfForceFontDownLoad = FALSE;
	gfpOut = stdout;
	gfTitleDisp = TRUE;
	gdwMagnification = 1000;
	gwDPI = 600;
	gnPrintPapers = 1;
	gnPdlType = DEFAULT_PDL;
	gfPrintEvenPage = gfPrintOddPage = FALSE;
	gnStartPage = 1; gnEndPage = 65535;
	gnOutPapers = 0;
	gnPaperFeedMode = 0;
	gnThresholdCharacters = LIPS_THRESHOLD_CHARACTERS;
	gnThresholdMemory = LIPS_THRESHOLD_MEMORY;
	pszAccountFile = pszRemoteUser = pszRemoteHost = NULL;
	SetupPdl();
	SetupColor();
	return TRUE;
}

static BOOL Setup(char *argv[])
{
#ifdef DVIPR_USE_KPATHSEA
	char	acEnvBuff[1024], *ptr;

	if ((ptr = getenv("PATH")) != NULL)
		strcpy(acEnvBuff, ptr);
	strcat(acEnvBuff, ":");
	strcat(acEnvBuff, MAKETEXPK_PATH);
	setenv("PATH", acEnvBuff, TRUE);
	kpse_set_progname(argv[0]);
	kpse_set_program_enabled( kpse_pk_format, 1, kpse_src_texmf_cnf );
#endif
	gfColorMode = (gnPdlType == PDL_LIPS4C || gnPdlType == PDL_CAPSL4) ? TRUE : FALSE;
	if (gpszWorkDirectory != NULL)
		chdir(gpszWorkDirectory);
	if (gnPdlType == PDL_LIPS3 || gnPdlType == PDL_CAPSL3) {
		gwDPI = 300;
#ifdef DVIPR_USE_KPATHSEA
		kpse_init_prog("DVIPDL", gwDPI, NULL, "cmr10");
#endif
	} else if (gnPdlType == PDL_LIPS4C) {
		gwDPI = 360;
#ifdef DVIPR_USE_KPATHSEA
		kpse_init_prog("DVIPDL", gwDPI, NULL, "cmr10");
#endif
	} else if (gnPdlType == PDL_LIPS4) {
		gwDPI = 600;
#ifdef DVIPR_USE_KPATHSEA
		kpse_init_prog("DVIPDL", gwDPI, NULL, "cmr10");
#endif
	}
	return TRUE;
}

static BOOL WriteAccount(void)
{
	time_t	JobTime;
	char	*ptr;
	FILE	*fp;

	if ((fp = fopen(pszAccountFile, "a")) == NULL)
		return FALSE;
	JobTime = time(0);
	ptr = ctime(&JobTime);
	*strchr(ptr, '\n') = '\0';
	fprintf(fp, "%s: dvipr[%d]: %s@%s TeX dvi file %d papers\n", ptr, getpid(), pszRemoteUser, pszRemoteHost, gnOutPapers * gnPrintPapers);
	fclose(fp);
	return TRUE;
}


static BOOL AnalizeArgument(int argc, char *argv[])
{
	char	*ptr, *ptr2;
	int		ct;

	gaszDviFileName[0] = '\0';
	for (ct = 1; ct < argc; ct++)
		if (*argv[ct] == '-') {
			switch (argv[ct][1]) {
				case 'b':
					gfTitleDisp = FALSE;
					break;
				case 'h':
					return FALSE;
					break;
				case 'D':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					if ((gwDPI = atoi(ptr)) < 0) {
						Error(ERR_INVALID_OPTION);
						return FALSE;
					}
					break;
				case 'l':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					if (!strcmp(ptr, "LIPS3"))
						gnPdlType = PDL_LIPS3;
					else if (!strcmp(ptr, "LIPS4C"))
						gnPdlType = PDL_LIPS4C;
					else if (!strcmp(ptr, "LIPS4"))
						gnPdlType = PDL_LIPS4;
					else if (!strcmp(ptr, "CAPSL3"))
						gnPdlType = PDL_CAPSL3;
					else if (!strcmp(ptr, "CAPSL4"))
						gnPdlType = PDL_CAPSL4;
					else {
						Error(ERR_INVALID_OPTION);
						return FALSE;
					}
					break;
				case 'e':
					gfPrintEvenPage = TRUE;
					break;
				case 'o':
					gfPrintOddPage = TRUE;
					break;
				case 'N':
					gfUseInternalFont = FALSE;
					break;
				case 'p':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					if ((gnStartPage = gnEndPage = atoi(strtok(ptr, ","))) < 1) {
						Error(ERR_BAD_OPTION);
						return FALSE;
					}
					if ((ptr2 = strtok(NULL, ",")) != NULL) {
						if (!(gnEndPage = atoi(ptr2)))
							gnEndPage = 65535;
						if (gnStartPage > gnEndPage) {
							Error(ERR_BAD_OPTION);
							return FALSE;
						}
					}
					break;
				case 'c':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					if ((gnPrintPapers = atoi(ptr)) < 0) {
						Error(ERR_INVALID_OPTION);
						return FALSE;
					}
					break;
				case 'f':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					if ((gnPaperFeedMode = atoi(ptr)) < 0 || 8 < atoi(ptr)) {
						Error(ERR_INVALID_OPTION);
						return FALSE;
					}
					break;
				case 'r':
					gfEnableFontDownLoad = FALSE;
					break;
				case 'R':
					gfForceFontDownLoad = TRUE;
					break;
				case 'm':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					if (atof(ptr) < 0. || 5. < atof(ptr)) {
						Error(ERR_INVALID_OPTION);
						return FALSE;
					}
					gdwMagnification = (DWORD)((double)gdwMagnification * pow(1.2, atof(ptr)));
					break;
				case 'M':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					if (atoi(ptr) < 0) {
						Error(ERR_INVALID_OPTION);
						return FALSE;
					}
					gdwMagnification = atoi(ptr);
					break;
				case 't':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					if (atoi(ptr) < 0 || 100 < atoi(ptr)) {
						Error(ERR_INVALID_OPTION);
						return FALSE;
					}
					gnThresholdCharacters = atoi(ptr);
					break;
				case 'T':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					if (atoi(ptr) < 0 || 100 < atoi(ptr)) {
						Error(ERR_INVALID_OPTION);
						return FALSE;
					}
					gnThresholdMemory = atoi(ptr);
					break;
				case 'w':
					if ((gpszWorkDirectory = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					break;
				case 'x':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					break;
				case 'y':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					break;
				case 'u':
					if ((pszRemoteUser = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					break;
				case 'U':
					if ((pszRemoteHost = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					break;
				case 'a':
					if ((pszAccountFile = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					break;
				case '\0':
					strcpy(gaszDviFileName, "-");
					break;
				default:
					return FALSE;
			}
		} else
			strncpy(gaszDviFileName, argv[ct], FILENAME_MAX);
	if (!strlen(gaszDviFileName)) {
		Error(ERR_DVI_NO_DVIFILE);
		return FALSE;
	}
	return TRUE;
}

static char *GetNextOption(int argc, char *argv[], int ct)
{
	if (!(++ct < argc)) {
fprintf(stderr, "op %d %d\n", argc, ct);
		Error(ERR_INVALID_OPTION);
		return NULL;
	}
	return argv[ct];
}
/*
static BOOL ReadEnvFile(void)
{
	FILE	*fp;

	if ((fp = fopen(DVIPR_ENVIRONMENT_FILE, "rt")) == NULL) {
		Error(ERR_ENVIRONMENTFILE_NOTFOUND);
		return FALSE;
	}
	fclose(fp);
	return TRUE;
}
*/
static void PrintTitle(void)
{
	fprintf(stderr, "dvipdl - TeX printer driver for LIPS-III/IV/IVc,CaPSL3/4 version %s\n", DVIPR_VERSION);
	fprintf(stderr, "This software is distributed by Kawaguti Laboratory of Fukui Univ. JAPAN\n\n");
}

static void PrintUsage(void)
{
	PrintTitle();
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "  -h        : Help message. (this)\n");
	fprintf(stderr, "  -T        : Suppress displaying title message.\n");
	fprintf(stderr, "  -l xxx    : Printer dependent language\n");
	fprintf(stderr, "            (e.g. LIPS3, LIPS4, LIPS4C)\n");
	fprintf(stderr, "  -p a[,[b]]: Print page from #a to #b\n");
	fprintf(stderr, "  -e        : Print even page\n");
	fprintf(stderr, "  -o        : Print odd page\n");
	fprintf(stderr, "  -c xxx    : Print n copies\n");
	fprintf(stderr, "  -f x      : Paper feed mode (0:Auto 1:Tray 2:Caset1, 3:Caset2 ... 8:Caset7)\n");
	fprintf(stderr, "  -D xxx    : DPI\n");
	fprintf(stderr, "  -m xxx    : magnification (scale = magstep xxx = 1.2^xxx)\n");
	fprintf(stderr, "  -M xxx    : magnification (scale = xxx/1000)\n");
	fprintf(stderr, "  -N        : Use DNP font (instead of internal font)\n");
	fprintf(stderr, "  -r        : Disable downloading font\n");
	fprintf(stderr, "  -R        : Force downloading font\n");
	fprintf(stderr, "  -t        : Threshold of characters for downloading fonts (Must be 0 to 100)\n");
	fprintf(stderr, "  -T        : Threshold of memory for downloading fonts (Must be 0 to 100)\n");
	fprintf(stderr, "  -         : Read .dvi file from standard-input\n");
}

void FatalExit(int nExit)
{
	exit(nExit);
}
