/*
 *  dvipr
 *      TeX Printer driver for The Canon LIPS III/IVc, CaPSL III printer
 *
 *      Developed at:
 *          Computer Systems Research Group
 *          Department of Information Science
 *          Fukui University
 *          9-1, Bunkyo-3, Fukui, 910 Japan
 *
 *      Authors:
 *          Yasumasa Koizumi, Atsushi Takagi, Kenichi Sugimoto,
 *          Yasuhide nya- Sawada, and Minato Kawaguti
 *
 *  The program may be copied and redistributed freely in the spirit of
 *  Free Software Foundation's "Copy Left", provided this title page is
 *  included in all of the source files involved.
 *
 *  Reports on porting to new platforms, bug fixes, etc., are highly
 *  appreciated. e-mail of this kind of communication shhould be addressedd
 *  to:
 *      dvipr@i1nws1.fukui-u.ac.jp
 */

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/param.h>
#include "newtypes.h"
#include "pdl.h"
#include "dvipr.h"
#include "dvipdl.h"
#include "dvi.h"
#include "file.h"
#include "gvars.h"
#include "error.h"
#include "funcdef.h"

#define	DVIPR_LPRCOMMAND_MAX	256

static char gaszDviFileName[FILENAME_MAX];
static char gaszDvipdlOptions[DVIPDL_OPTIONS_MAX];
static char gaszLprCommand[DVIPR_LPRCOMMAND_MAX];
static BOOL	gfDisplayTitle;

static BOOL Initialize(char *argv[]);
static BOOL AnalizeArgument(int argc, char *argv[]);
static char *GetNextOption(int argc, char *argv[], int ct);
static void PrintTitle(void);
static void PrintUsage(void);

int main(int argc, char *argv[])
{
	int		ch;
	char	acWorkDirectory[MAXPATHLEN];
	FILE	*fp, *pp;

	if (!Initialize(argv))
		return -1;
	if (!AnalizeArgument(argc, argv)) {
		PrintUsage();
		return -2;
	}
	strcat(gaszDvipdlOptions, " -w ");
	strcat(gaszDvipdlOptions, getwd(acWorkDirectory));
	if (gfDisplayTitle)
		PrintTitle();
	if ((fp = fopen(gaszDviFileName, "rb")) == NULL) {
		strcat(gaszDviFileName, ".dvi");
		if ((fp = fopen(gaszDviFileName, "rb")) == NULL) {
			Error(ERR_DVI_NOT_FOUND);
			return -3;
		}
	}
	if ((pp = popen(gaszLprCommand, "w")) == NULL) {
		printf("Can't execute %s.", DVIPR_LPR_CMD);
		return -1;
	}
	FpWrite4(pp, DVIPR_ID);
	FpWrite2(pp, strlen(gaszDvipdlOptions) + 1);
	fwrite(gaszDvipdlOptions, strlen(gaszDvipdlOptions) + 1, 1, pp);
	while ((ch = fgetc(fp)) != EOF)
		fputc(ch, pp);
	pclose(pp);
	fclose(fp);
	return 0;
}

static BOOL Initialize(char *argv[])
{
	gfDisplayTitle = TRUE;
	memset(gaszDviFileName, 0, FILENAME_MAX);
	memset(gaszDvipdlOptions, 0, DVIPDL_OPTIONS_MAX);
	return TRUE;
}

static void PrintTitle(void)
{
	printf("dvipr - TeX printer driver for LIPS-III/IV/IVc,CaPSL3/4 version %s\n", DVIPR_VERSION);
	printf("This software is distributed by Kawaguti Laboratory of Fukui Univ. JAPAN\n\n");
}

static void PrintUsage(void)
{
	PrintTitle();
	printf("Options:\n");
	printf("  -h        : Help message (this)\n");
	printf("  -Pxxx     : Output printer name\n");
	printf("  -l xxx    : Printer language\n");
	printf("            (e.g. LIPS3, LIPS4, LIPS4C)\n");
	printf("  -p a[,[b]]: Print page from #a to #b\n");
	printf("  -e        : Print even page\n");
	printf("  -o        : Print odd page\n");
	printf("  -c xxx    : Print n copies\n");
	printf("  -f x      : Paper feed mode (0:Auto 1:Tray 2:Caset1, 3:Caset2 ... 8:Caset7)\n");
	printf("  -D xxx    : DPI\n");
	printf("  -m xxx    : magnification (scale = magstep xxx = 1.2^xxx)\n");
	printf("  -M xxx    : magnification (scale = xxx/1000)\n");
	printf("  -N        : Use DNP font (instead of internal font)\n");
	printf("  -r        : Disable downloading fonts\n");
	printf("  -R        : Force downloading fonts\n");
	printf("  -t xxx    : Threshold of characters for downloading fonts (Must be 0 to 100)\n");
	printf("  -T xxx    : Threshold of build-in memory for downloading fonts (Must be 0 to 100)\n");
	printf("  -d        : DEBUG option (return mail about your job)\n");
}

static BOOL AnalizeArgument(int argc, char *argv[])
{
	char	*ptr, *ptr2;
	int		ct, wk;

	sprintf(gaszLprCommand, "%s %s", DVIPR_LPR_CMD, DVIPR_LPR_ADDITIONAL);
	for (ct = 1; ct < argc; ct++)
		if (*argv[ct] == '-') {
			switch (argv[ct][1]) {
				case 'b':
					gfDisplayTitle = FALSE;
					break;
				case 'h':
					return FALSE;
					break;
				case 'P':
					strcat(gaszLprCommand, " -P");
					strcat(gaszLprCommand, &argv[ct][2]);
					break;
				case 'D':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					if ((gwDPI = atoi(ptr)) < 0) {
						Error(ERR_BAD_OPTION);
						return FALSE;
					}
					strcat(gaszDvipdlOptions, " -D ");
					strcat(gaszDvipdlOptions, ptr);
					break;
				case 'l':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					if (!strcmp(ptr, "LIPS3") || !strcmp(ptr, "LIPS4") || !strcmp(ptr, "LIPS4C") || !strcmp(ptr, "CAPSL3") || !strcmp(ptr, "CAPSL4")) {
						strcat(gaszDvipdlOptions, " -l ");
						strcat(gaszDvipdlOptions, ptr);
					} else {
						Error(ERR_BAD_OPTION);
						return FALSE;
					}
					break;
				case 'e':
					strcat(gaszDvipdlOptions, " -e ");
					break;
				case 'o':
					strcat(gaszDvipdlOptions, " -o ");
					break;
				case 'p':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					strcat(gaszDvipdlOptions, " -p ");
					strcat(gaszDvipdlOptions, ptr);
					wk = atoi(strtok(ptr, ","));
					if ((ptr2 = strtok(NULL, ",")) != NULL)
						if (wk > atoi(ptr2) && atoi(ptr2)) {
							Error(ERR_BAD_OPTION);
							return FALSE;
						}
					break;
				case 'c':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					strcat(gaszDvipdlOptions, " -c ");
					strcat(gaszDvipdlOptions, ptr);
					if (atoi(ptr) < 0) {
						Error(ERR_BAD_OPTION);
						return FALSE;
					}
					break;
				case 'f':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					strcat(gaszDvipdlOptions, " -f ");
					strcat(gaszDvipdlOptions, ptr);
					if (atoi(ptr) < 0 || 8 < atoi(ptr)) {
						Error(ERR_BAD_OPTION);
						return FALSE;
					}
					break;
				case 'm':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					strcat(gaszDvipdlOptions, " -m ");
					strcat(gaszDvipdlOptions, ptr);
					if (atoi(ptr) < 0 || 5 < atoi(ptr)) {
						Error(ERR_BAD_OPTION);
						return FALSE;
					}
					break;
				case 'M':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					strcat(gaszDvipdlOptions, " -M ");
					strcat(gaszDvipdlOptions, ptr);
					if (atoi(ptr) < 0) {
						Error(ERR_BAD_OPTION);
						return FALSE;
					}
					break;
				case 'N':
					strcat(gaszDvipdlOptions, " -N ");
					break;
				case 'r':
					strcat(gaszDvipdlOptions, " -r ");
					break;
				case 'R':
					strcat(gaszDvipdlOptions, " -R ");
					break;
				case 't':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					strcat(gaszDvipdlOptions, " -t ");
					strcat(gaszDvipdlOptions, ptr);
					if (atoi(ptr) < 0 || 100 < atoi(ptr)) {
						Error(ERR_BAD_OPTION);
						return FALSE;
					}
					break;
				case 'T':
					if ((ptr = GetNextOption(argc, argv, ct++)) == NULL)
						return FALSE;
					strcat(gaszDvipdlOptions, " -T ");
					strcat(gaszDvipdlOptions, ptr);
					if (atoi(ptr) < 0 || 100 < atoi(ptr)) {
						Error(ERR_BAD_OPTION);
						return FALSE;
					}
					break;
				case 'd':
					strcat(gaszLprCommand, " -m");
					break;
				default:
					return FALSE;
			}
		} else
			strncpy(gaszDviFileName, argv[ct], FILENAME_MAX);
	if (!strlen(gaszDviFileName)) {
		Error(ERR_DVI_NOTSPECIFIED);
		return FALSE;
	}
	return TRUE;
}

static char *GetNextOption(int argc, char *argv[], int ct)
{
	if (!(++ct < argc)) {
		Error(ERR_BAD_OPTION);
		return NULL;
	}
	return argv[ct];
}

void FatalExit(int nExit)
{
	exit(nExit);
}
