/*
 *  dvipr
 *      TeX Printer driver for The Canon LIPS III/IVc, CaPSL III printer
 *
 *      Developed at:
 *          Computer Systems Research Group
 *          Department of Information Science
 *          Fukui University
 *          9-1, Bunkyo-3, Fukui, 910 Japan
 *
 *      Authors:
 *          Yasumasa Koizumi, Atsushi Takagi, Kenichi Sugimoto,
 *          and Minato Kawaguti
 *
 *  The program may be copied and redistributed freely in the spirit of
 *  Free Software Foundation's "Copy Left", provided this title page is
 *  included in all of the source files involved.
 *
 *  Reports on porting to new platforms, bug fixes, etc., are highly
 *  appreciated. e-mail of this kind of communication shhould be addressedd
 *  to:
 *      dvipr@i1nws1.fukui-u.ac.jp
 * ------------------------------------------------------------------------
 * <<< TechNote >>>
 *  fnum is used by following rule.
 *     0 - 255: normal font number
 *   256 - 999: virtual font number assigned to DNP font
 *  1024 -    : virtual font number assigned to TATEGUMI (=YOKOGUMI fnum + 1024)
 * ------------------------------------------------------------------------
 */

#include <stdlib.h>
#include <memory.h>
#include "font.h"
#include "pk.h"
#include "gvars.h"
#include "pdl.h"
#include "dvipdl.h"
#include "newtypes.h"
#include "funcdef.h"

BOOL SelectFont(Fonts* font)
{
	if (gpfnPdlSelectFont(font))
		return TRUE;
	if (ReadPK(font))
		return TRUE;
	return FALSE;
}

Fonts* GetFonts(unsigned DWORD fnum)
{
	Fonts* ptr;

	for (ptr = pFontRoot; ptr; ptr = ptr->next)
		if (ptr->dwFontNum == fnum)
			return ptr;
fprintf(stderr, "GetFonts: font not found[%d]\n", fnum);
abort();
	Error(ERR_DVI_INVALID_FORMAT);
	return NULL;
}

unsigned DWORD GetFontNumber(Fonts *pFont)
{
	Fonts* ptr;

	for (ptr = pFontRoot; ptr; ptr = ptr->next) {
		if (strcmp(ptr->acName, pFont->acName))
			continue;
		if (ptr->dwScaled != pFont->dwScaled)
			continue;
		if (ptr->dwDesignSize != pFont->dwDesignSize)
			continue;
		return ptr->dwFontNum;
	}
	return 0;
}

unsigned DWORD GetFontWidthMax(Fonts *pFont)
{
	unsigned DWORD	dwWidthMax;
	Gryph			*pGryph;

	dwWidthMax = 0;
	for (pGryph = pFont->pGryph; pGryph != NULL; pGryph = pGryph->next)
		if (!dwWidthMax || dwWidthMax < pGryph->dwWidth)
			dwWidthMax = pGryph->dwWidth;
	return dwWidthMax;
}

unsigned DWORD GetFontHeightMax(Fonts *pFont)
{
	unsigned DWORD	dwHeightMax;
	Gryph			*pGryph;

	dwHeightMax = 0;
	for (pGryph = pFont->pGryph; pGryph != NULL; pGryph = pGryph->next)
		if (!dwHeightMax || dwHeightMax < pGryph->dwHeight)
			dwHeightMax = pGryph->dwHeight;
	return dwHeightMax;
}

unsigned DWORD GetFontGryphs(Fonts *pFont)
{
	unsigned WORD	ct;
	Gryph			*pGryph;

	ct = 0;
	for (pGryph = pFont->pGryph; pGryph != NULL; pGryph = pGryph->next)
		ct++;
	return ct;
}

void DebugFontList(void)
{
	Fonts* ptr;

	fprintf(stderr, "-----\n");
	for (ptr = pFontRoot; ptr; ptr = ptr->next)
		fprintf(stderr, "%s[%d] %d %d\n", ptr->acName, ptr->dwFontNum, ptr->dwScaled, ptr->dwDesignSize);
	fprintf(stderr, "-----\n");
}

void ResetCounter(void)
{
	Fonts	*pFont;
	Gryph	*pGryph;


	for (pFont = pFontRoot; pFont; pFont = pFont->next)
		for (pGryph = pFont->pGryph; pGryph; pGryph = pGryph->next)
			pGryph->dwCounter = 0;
}

Gryph* GetGryph(Fonts* font, unsigned DWORD code)
{
	Gryph* ptr;

	for (ptr = font->pGryph; ptr; ptr = ptr->next)
		if (ptr->dwCharCode == code)
			return ptr;
fprintf(stderr, "Error: GetGryph cc=%d, fname=%s\n", code, font->acName);
	Error(ERR_DVI_INVALID_FORMAT);
	return NULL;
}

unsigned DWORD GetCharWidth(Fonts* font, unsigned DWORD code)
{
	CharWidth*		ptr;
	BOOL			bDefaultValid;
	unsigned DWORD	dwDefaultCharWidth = 0;

	bDefaultValid = FALSE;
	for (ptr = font->pCharWidth; ptr; ptr = ptr->next) {
		if (!(ptr->dwCharCode)) {
			bDefaultValid = TRUE;
			dwDefaultCharWidth = ptr->dwWidth;
		}
		if (ptr->dwCharCode == code)
			return ptr->dwWidth;
	}
	if (!bDefaultValid) {
		Error(ERR_INTERNALERROR);
		FatalExit(-1);
	}
	return dwDefaultCharWidth;
}

Fonts* CreateFont(void)
{
	Fonts *pNewFont, *ptr;

	if ((pNewFont = (Fonts*)malloc(sizeof (Fonts))) == NULL) {
		Error(ERR_NOMEMORY);
		return NULL;
	}
	memset(pNewFont, 0, sizeof (Fonts));
	pNewFont->next = NULL;
	if (pFontRoot == NULL)
		pFontRoot = pNewFont;
	else {
		for (ptr = pFontRoot; ptr->next; ptr = ptr->next);
		ptr->next = pNewFont;
	}
	return pNewFont;
}

void CopyFont(Fonts *dest, Fonts *src)
{
	Fonts*	next;

	next = dest->next;
	memcpy(dest, src, sizeof (Fonts));
	dest->next = next;
}

Gryph* CreateGryph(Fonts* pFont)
{
	Gryph *pNewGryph, *ptr;

	if ((pNewGryph = (Gryph*)malloc(sizeof (Gryph))) == NULL) {
		Error(ERR_NOMEMORY);
		return NULL;
	}
	memset(pNewGryph, 0, sizeof (Gryph));
	pNewGryph->next = NULL;
	if (pFont->pGryph == NULL)
		pFont->pGryph = pNewGryph;
	else {
		for (ptr = pFont->pGryph; ptr->next; ptr = ptr->next);
		ptr->next = pNewGryph;
	}
	return pNewGryph;
}

CharWidth* CreateCharWidth(Fonts* pFont)
{
	CharWidth *pNewCharWidth, *ptr;

	if ((pNewCharWidth = (CharWidth*)malloc(sizeof (CharWidth))) == NULL) {
		Error(ERR_NOMEMORY);
		return NULL;
	}
	memset(pNewCharWidth, 0, sizeof (CharWidth));
	pNewCharWidth->next = NULL;
	if (pFont->pCharWidth == NULL)
		pFont->pCharWidth = pNewCharWidth;
	else {
		for (ptr = pFont->pCharWidth; ptr->next; ptr = ptr->next);
		ptr->next = pNewCharWidth;
	}
	return pNewCharWidth;
}

BOOL RotateFont(Fonts *pFont)
{
	int		x, y;
	unsigned char	cache = 0;
	unsigned char	*pNewBitmap, *ptr;
	Gryph	*pGryph;

	for (pGryph = pFont->pGryph; pGryph != NULL; pGryph = pGryph->next) {
		if ((ptr = pNewBitmap = (unsigned char *)malloc(round8(pGryph->dwHeight) * pGryph->dwWidth)) == NULL) {
			Error(ERR_NOMEMORY);
			return FALSE;
		}
		for (x = pGryph->dwWidth - 1; x >= 0; x--) {
			for (y = 0; y < pGryph->dwHeight; y++) {
				cache <<= 1;
				if (pGryph->pBits[x / 8 + y * round8(pGryph->dwWidth)] & ((unsigned char)0x80 >> (x % 8)))
					cache |= 0x01;
				if ((y % 8) == 7)
					*pNewBitmap++ = cache;
			}
			if (y % 8) {
				cache <<= (8 - (y % 8));
				*pNewBitmap++ = cache;
			}
		}
		free(pGryph->pBits);
		pGryph->pBits = ptr;
	}
	return TRUE;
}
