/*
 *  dvipr
 *      TeX Printer driver for The Canon LIPS III/IVc, CaPSL III printer
 *
 *      Developed at:
 *          Computer Systems Research Group
 *          Department of Information Science
 *          Fukui University
 *          9-1, Bunkyo-3, Fukui, 910 Japan
 *
 *      Authors:
 *          Yasumasa Koizumi, Atsushi Takagi, Kenichi Sugimoto,
 *          and Minato Kawaguti
 *
 *  The program may be copied and redistributed freely in the spirit of
 *  Free Software Foundation's "Copy Left", provided this title page is
 *  included in all of the source files involved.
 *
 *  Reports on porting to new platforms, bug fixes, etc., are highly
 *  appreciated. e-mail of this kind of communication shhould be addressedd
 *  to:
 *      dvipr@i1nws1.fukui-u.ac.jp
 */

#include <string.h>
#include "font.h"
#include "tfm.h"
#include "dvi.h"
#include "error.h"
#include "dvipdl.h"
#include "dnp.h"
#include "pdl.h"
#include "funcdef.h"

static int		nFontNumber = 256;

static unsigned DWORD PutInternalChar(unsigned DWORD dwCharCode, unsigned DWORD fnum);

BOOL SelectJapaneseFont(unsigned DWORD jis, Fonts *pCurrentFont)
{
	int				nCode, nFontID;
	unsigned WORD	kuten;
	char			*pFontPrefix;
	Fonts			font, *pFont;

	if (gfUseInternalFont)
		return TRUE;
	if (strncmp(pCurrentFont->acName, "min", 3) && strncmp(pCurrentFont->acName, "goth", 4) && strncmp(pCurrentFont->acName, "tmin", 4) && strncmp(pCurrentFont->acName, "tgoth", 5))
		return TRUE;
	kuten = JisToKuten(jis);
	GetDnpFontInfoByKutenCode(kuten, &pFontPrefix, &nCode, &nFontID);
	if (!strncmp(pCurrentFont->acName, "min", 3) || !strncmp(pCurrentFont->acName, "tmin", 4))
		sprintf(font.acName, "dmj%s%d", pFontPrefix, (int)((double)pCurrentFont->dwDesignSize / (double)Pow(2., 16)));
	else if (!strncmp(pCurrentFont->acName, "goth", 4) || !strncmp(pCurrentFont->acName, "tgoth", 5))
		sprintf(font.acName, "dgj%s%d", pFontPrefix, (int)((double)pCurrentFont->dwDesignSize / (double)Pow(2., 16)));
	else {
		Error2(ERR_INTERNALERROR, pCurrentFont->acName);
		FatalExit(-1);
	}
	font.dwScaled = pCurrentFont->dwScaled;
	font.dwDesignSize = pCurrentFont->dwDesignSize;
	if (GetFontNumber(&font))
		return TRUE;
	pFont = CreateFont();
	pFont->dwFontNum = nFontNumber++;
	pFont->dwCheckSum = 0;
	pFont->dwScaled = font.dwScaled;
	pFont->dwDesignSize = font.dwDesignSize;
	memset(pFont->acDir, 0, FILENAME_MAX);
	strcpy(pFont->acName, font.acName);
	if (!SelectFont(pFont))
		return FALSE;
	return TRUE;
}

unsigned DWORD PutChar(unsigned DWORD dwCharCode, unsigned DWORD fnum)
{
	int				nCode, nFontID;
	unsigned WORD	kuten;
	char			*pFontPrefix;
	Fonts			font, *pCurrentFont;

/*fprintf(stderr, "PutChar code=%d, fnum=%d\n", dwCharCode, fnum);*/
	if (gfUseInternalFont)
		return PutInternalChar(dwCharCode, fnum);
	if (dwCharCode < 0x100)
		return gpfnPdlPutChar(dwCharCode);
	kuten = JisToKuten(dwCharCode);
	GetDnpFontInfoByKutenCode(kuten, &pFontPrefix, &nCode, &nFontID);
	pCurrentFont = GetFonts(fnum);
	if (!strncmp(pCurrentFont->acName, "min", 3) || !strncmp(pCurrentFont->acName, "tmin", 4))
		sprintf(font.acName, "dmj%s%d", pFontPrefix, (int)((double)pCurrentFont->dwDesignSize / (double)Pow(2., 16)));
	else if (!strncmp(pCurrentFont->acName, "goth", 4) || !strncmp(pCurrentFont->acName, "tgoth", 5))
		sprintf(font.acName, "dgj%s%d", pFontPrefix, (int)((double)pCurrentFont->dwDesignSize / (double)Pow(2., 16)));
	else {
fprintf(stderr, "PutChar\n");
		Error2(ERR_INTERNALERROR, pCurrentFont->acName);
		FatalExit(-1);
	}
	font.dwScaled = pCurrentFont->dwScaled;
	font.dwDesignSize = pCurrentFont->dwDesignSize;
	gpfnPdlChangeFont(GetFontNumber(&font));
	gpfnPdlPutChar(nCode);						
	return (DWORD)((double)GetCharWidth(pCurrentFont, dwCharCode) / Pow(2., 20) * (double)pCurrentFont->dwScaled);
}

static unsigned DWORD PutInternalChar(unsigned DWORD dwCharCode, unsigned DWORD fnum)
{
	int	ff;
	double			lfDesignSizePt;
	Fonts			*pCurrentFont, font;
	unsigned WORD	kuten;
	
/*fprintf(stderr, "PIC code=%d, font=%d\n", dwCharCode, fnum);*/
	pCurrentFont = GetFonts(fnum);
	if (strncmp(pCurrentFont->acName, "dmj", 3) && strncmp(pCurrentFont->acName, "dgj", 3))
		return gpfnPdlPutChar(dwCharCode);
	GetKutenCodeByDnpFontInfo(&(pCurrentFont->acName[3]), dwCharCode, &kuten);
	CopyFont(&font, pCurrentFont);
	lfDesignSizePt = (double)pCurrentFont->dwDesignSize / (double)Pow(2., 16);
	if (lfDesignSizePt > 10) {
		sprintf(font.acName, "%s10", strncmp(pCurrentFont->acName, "dmj", 3) ? "goth" : "min");
		font.dwScaled *= lfDesignSizePt / 10.;
		font.dwDesignSize = 10 * Pow(2., 16);
	} else
		sprintf(font.acName, "%s%d", strncmp(pCurrentFont->acName, "dmj", 3) ? "goth" : "min", (int)lfDesignSizePt);
/*fprintf(stderr, "search: font=%s scale=%d design=%d\n", font.acName, font.dwScaled, font.dwDesignSize);*/
	ff = GetFontNumber(&font);
/*fprintf(stderr, "cf=%d\n", ff);*/
	gpfnPdlChangeFont(GetFontNumber(&font));
	gpfnPdlPutChar(KutenToJis(kuten));
	return (DWORD)((double)GetCharWidth(pCurrentFont, dwCharCode) / Pow(2., 20) * (double)pCurrentFont->dwScaled);
}

BOOL ReplaceJapaneseFont(Fonts *pFont)
{
	Fonts			font, *pNewFont;
	double			lfDesignSizePt;

	if (strncmp(pFont->acName, "dmj", 3) && strncmp(pFont->acName, "dgj", 3))
		return FALSE;
	ReadTFM(pFont);
	CopyFont(&font, pFont);
	if ((lfDesignSizePt = (double)pFont->dwDesignSize / (double)Pow(2., 16)) > 10.) {
		sprintf(font.acName, "%s10", strncmp(pFont->acName, "dmj", 3) ? "goth" : "min");
		font.dwScaled *= lfDesignSizePt / 10.;
		font.dwDesignSize = 10 * Pow(2., 16);
	} else
		sprintf(font.acName, "%s%d", strncmp(pFont->acName, "dmj", 3) ? "goth" : "min", (int)lfDesignSizePt);
	if (!GetFontNumber(&font)) {
		pNewFont = CreateFont();
		CopyFont(pNewFont, &font);
		pNewFont->dwFontNum = nFontNumber++;
		SelectFont(pNewFont);
		pNewFont->Type = FONT_DNPINTERNAL;
	}
	return TRUE;
}
