/*
 *  dvipr
 *      TeX Printer driver for The Canon LIPS III/IVc, CaPSL III printer
 *
 *      Developed at:
 *          Computer Systems Research Group
 *          Department of Information Science
 *          Fukui University
 *          9-1, Bunkyo-3, Fukui, 910 Japan
 *
 *      Authors:
 *          Yasumasa Koizumi, Atsushi Takagi, Kenichi Sugimoto,
 *          and Minato Kawaguti
 *
 *  The program may be copied and redistributed freely in the spirit of
 *  Free Software Foundation's "Copy Left", provided this title page is
 *  included in all of the source files involved.
 *
 *  Reports on porting to new platforms, bug fixes, etc., are highly
 *  appreciated. e-mail of this kind of communication shhould be addressedd
 *  to:
 *      dvipr@i1nws1.fukui-u.ac.jp
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "newtypes.h"
#include "site.h"
#include "dvi.h"
#include "pdl.h"
#include "dvipdl.h"
#include "lips.h"
#include "gs.h"
#include "error.h"
#include "funcdef.h"

#define	PSfile_TOKEN_STR		" \t"
#define	epsfile_TOKEN_STR		" \t"
#define	postscriptbox_TOKEN_STR	"{}"

static BOOL CreateGsPipe(FILE *fp[2], char *pGsDevice);
static BOOL GetBoundingBox(char *pszFile, double *llx, double *lly, double *urx, double *ury);

BOOL SpecialPSfile(char *pszSpecial, char *pGsDevice)
{
	int		ch;
	char	*pszPSFile, *ptr, acBuff[BUFSIZ];
	double	llx, lly, urx, ury, rwi, rhi, hscale, vscale;
	FILE	*pp[2], *fp;

	pszPSFile = NULL;
	llx = lly = urx = ury = rwi = rhi = 0.;
	hscale = vscale = 1.;
	for (ptr = strtok(pszSpecial, PSfile_TOKEN_STR); ptr; ptr = strtok(NULL, PSfile_TOKEN_STR)) {
		if (!strncmp(ptr, "PSfile", 6))
			pszPSFile = &ptr[7];
		else if (!strncmp(ptr, "llx", 3))
			llx = atof(&ptr[4]);
		else if (!strncmp(ptr, "lly", 3))
			lly = atof(&ptr[4]);
		else if (!strncmp(ptr, "urx", 3))
			urx = atof(&ptr[4]);
		else if (!strncmp(ptr, "ury", 3))
			ury = atof(&ptr[4]);
		else if (!strncmp(ptr, "rwi", 3))
			rwi = atof(&ptr[4]);
		else if (!strncmp(ptr, "rhi", 3))
			rhi = atof(&ptr[4]);
		else
			Error2(ERR_GS_PSFILE_UNKNOWN_PARAMETER, ptr);
	}
	if (pszPSFile == NULL) {
		Error(ERR_GS_PSFILE_INVALID);
		return FALSE;
	}
	if (*pszPSFile == '\"') {
		pszPSFile++;
		pszPSFile[strlen(pszPSFile) - 1] = '\0';
	}
	if (!CreateGsPipe(pp, pGsDevice))
		return FALSE;
	if ((fp = fopen(pszPSFile, "rt")) == NULL) {
		Error2(ERR_GS_PSFILE_NOTFOUND, pszPSFile);
		return FALSE;
	}
	if (rwi)
		hscale = vscale = rwi / 10. / (urx - llx);
	if (rhi)
		vscale = rhi / 10. / (urx - llx);
	fprintf(pp[1], "%f %f translate\n", SpToHPSp(xDviReg.h) - hscale * llx, (841. - SpToVPSp(xDviReg.v)) - vscale * lly);
	fprintf(pp[1], "%f %f scale\n", hscale, vscale);
	fprintf(pp[1], "newpath\n");
	fprintf(pp[1], "%f %f moveto\n", urx, ury);
	fprintf(pp[1], "%f %f lineto\n", llx, ury);
	fprintf(pp[1], "%f %f lineto\n", llx, lly);
	fprintf(pp[1], "%f %f lineto\n", urx, lly);
	fprintf(pp[1], "closepath\nclip\n");
	while (fgets(acBuff, BUFSIZ, fp) != NULL)
		fputs(acBuff, pp[1]);
	fprintf(pp[1], "showpage\nquit\n");
	fflush(pp[1]);
	fclose(fp);
	while ((ch = fgetc(pp[0])) != EOF)
		gpfnPdlOutChar(ch);
	return TRUE;
}

BOOL Specialepsfile(char *pszSpecial, char *pGsDevice)
{
	int		ch;
	char	*pszepsfile, *ptr, acBuff[BUFSIZ];
	double	llx, lly, urx, ury, hscale, vscale, hsize, vsize;
	FILE	*pp[2], *fp;

	pszepsfile = NULL;
	hscale = vscale = hsize = vsize = 0.;
	for (ptr = strtok(pszSpecial, epsfile_TOKEN_STR); ptr; ptr = strtok(NULL, epsfile_TOKEN_STR)) {
		if (!strncmp(ptr, "epsfile", 7))
			pszepsfile = &ptr[8];
		else if (!strncmp(ptr, "hsize", 5))
			hsize = atof(&ptr[6]);
		else if (!strncmp(ptr, "vsize", 5))
			vsize = atof(&ptr[6]);
		else if (!strncmp(ptr, "hscale", 6))
			hscale = atof(&ptr[7]);
		else if (!strncmp(ptr, "vscale", 6))
			vscale = atof(&ptr[7]);
		else
			Error2(ERR_GS_PSFILE_UNKNOWN_PARAMETER, ptr);
	}
	if (pszepsfile == NULL) {
		Error(ERR_GS_PSFILE_INVALID);
		return FALSE;
	}
	if (!GetBoundingBox(pszepsfile, &llx, &lly, &urx, &ury)) {
		Error2(ERR_GS_PSFILE_CANTOPEN, pszepsfile);
		return FALSE;
	}
	if (!CreateGsPipe(pp, pGsDevice))
		return FALSE;
	if ((fp = fopen(pszepsfile, "rt")) == NULL) {
		Error2(ERR_GS_PSFILE_NOTFOUND, pszepsfile);
		return FALSE;
	}
	if (!hscale)
		hscale = hsize / (urx - llx);
	if (!vscale)
		vscale = vsize / (ury - lly);
	fprintf(pp[1], "%f %f translate\n", SpToHPSp(xDviReg.h) - hscale * llx, (841. - SpToVPSp(xDviReg.v)) - vscale * ury);
	fprintf(pp[1], "%f %f scale\n", hscale, vscale);
	fprintf(pp[1], "newpath\n");
	fprintf(pp[1], "%f %f moveto\n", urx, ury);
	fprintf(pp[1], "%f %f lineto\n", llx, ury);
	fprintf(pp[1], "%f %f lineto\n", llx, lly);
	fprintf(pp[1], "%f %f lineto\n", urx, lly);
	fprintf(pp[1], "closepath\nclip\n");
	while (fgets(acBuff, BUFSIZ, fp) != NULL)
		fputs(acBuff, pp[1]);
	fprintf(pp[1], "showpage\nquit\n");
	fflush(pp[1]);
	fclose(fp);
	while ((ch = fgetc(pp[0])) != EOF)
		gpfnPdlOutChar(ch);
	return TRUE;
}

BOOL Specialpostscriptbox(char *pszSpecial, char *pGsDevice)
{
	int		ch;
	char	*pszpostscriptbox, acBuff[BUFSIZ];
	double	llx, lly, urx, ury, width, height, hscale, vscale;
	FILE	*pp[2], *fp;

	llx = lly = urx = ury = 0.;
	hscale = vscale = 1.;
	if (sscanf(pszSpecial, "postscriptbox{%lf pt}{%lf pt}", &width, &height) != 2) {
		Error(ERR_GS_PSFILE_INVALID);
		return FALSE;
	}
	strtok(pszSpecial, postscriptbox_TOKEN_STR);
	strtok(NULL, postscriptbox_TOKEN_STR); strtok(NULL, postscriptbox_TOKEN_STR);
	pszpostscriptbox = strtok(NULL, postscriptbox_TOKEN_STR);
	if (!GetBoundingBox(pszpostscriptbox, &llx, &lly, &urx, &ury)) {
		Error2(ERR_GS_PSFILE_CANTOPEN, pszpostscriptbox);
		return FALSE;
	}
	if (!CreateGsPipe(pp, pGsDevice))
		return FALSE;
	if ((fp = fopen(pszpostscriptbox, "rt")) == NULL) {
		Error2(ERR_GS_PSFILE_NOTFOUND, pszpostscriptbox);
		return FALSE;
	}
	hscale = width / (urx - llx);
	vscale = height / (ury - lly);
	fprintf(pp[1], "%f %f translate\n", SpToHPSp(xDviReg.h) - hscale * llx, (841. - SpToVPSp(xDviReg.v)) - vscale * lly);
	fprintf(pp[1], "%f %f scale\n", hscale, vscale);
	fprintf(pp[1], "newpath\n");
	fprintf(pp[1], "%f %f moveto\n", urx, ury);
	fprintf(pp[1], "%f %f lineto\n", llx, ury);
	fprintf(pp[1], "%f %f lineto\n", llx, lly);
	fprintf(pp[1], "%f %f lineto\n", urx, lly);
	fprintf(pp[1], "closepath\nclip\n");
	while (fgets(acBuff, BUFSIZ, fp) != NULL)
		fputs(acBuff, pp[1]);
	fprintf(pp[1], "showpage\nquit\n");
	fflush(pp[1]);
	fclose(fp);
	while ((ch = fgetc(pp[0])) != EOF)
		gpfnPdlOutChar(ch);
	return TRUE;
}

BOOL SpecialRawPS(char *pszSpecial, char *pGsDevice)
{
	int		ch;
	FILE	*pp[2];

	if (!CreateGsPipe(pp, pGsDevice))
		return FALSE;
	fprintf(pp[1], "%f %f translate\n", SpToHPSp(xDviReg.h), 841. - SpToVPSp(xDviReg.v));
	fputs("%!PS-Adobe-2.0\n", pp[1]);
	fputs(&pszSpecial[1], pp[1]);
	fprintf(pp[1], "\nshowpage\nquit\n");
	fflush(pp[1]);
	while ((ch = fgetc(pp[0])) != EOF)
		gpfnPdlOutChar(ch);
	return TRUE;
}

static BOOL CreateGsPipe(FILE *fp[2], char *pGsDevice)
{
	int		fdParent[2], fdChild[2];
	char	acDeviceParam[32];

	if (pipe(fdParent) == -1 || pipe(fdChild) == -1)
		return FALSE;
	if (fork()) {
		close(fdParent[1]);
		close(fdChild[0]);
		if ((fp[0] = fdopen(fdParent[0], "r")) == NULL)
			Error(ERR_INTERNALERROR);
		if ((fp[1] = fdopen(fdChild[1], "w")) == NULL)
			Error(ERR_INTERNALERROR);
	} else {
		close(fdParent[0]);
		close(fdChild[1]);

		close(STDOUT_FILENO);
		dup(fdParent[1]);
		close(STDIN_FILENO);
		dup(fdChild[0]);

		close(fdParent[1]);
		close(fdChild[0]);
		sprintf(acDeviceParam, "-sDEVICE=%s", pGsDevice);
		execl(DVIPR_GHOSTSCRIPT_CMD, "gs", "-q", acDeviceParam, "-sOUTPUTFILE=-", "-dNOPAUSE", "-", NULL);
	}
	return TRUE;
}

static BOOL GetBoundingBox(char *pszFile, double *llx, double *lly, double *urx, double *ury)
{
	char	buff[BUFSIZ];
	FILE	*fp;

	if ((fp = fopen(pszFile, "rt")) == NULL)
		return FALSE;
	while (fgets(buff, BUFSIZ, fp) != NULL)
		if (!strncmp(buff, "%%BoundingBox:", 14)) {
			sscanf(&buff[14], "%lf %lf %lf %lf", llx, lly, urx, ury);
			break;
		}
	fclose(fp);
	return TRUE;
}
