/*
 *  dvipr
 *      TeX Printer driver for The Canon LIPS III/IVc, CaPSL III printer
 *
 *      Developed at:
 *          Computer Systems Research Group
 *          Department of Information Science
 *          Fukui University
 *          9-1, Bunkyo-3, Fukui, 910 Japan
 *
 *      Authors:
 *          Yasumasa Koizumi, Atsushi Takagi, Kenichi Sugimoto,
 *          and Minato Kawaguti
 *
 *  The program may be copied and redistributed freely in the spirit of
 *  Free Software Foundation's "Copy Left", provided this title page is
 *  included in all of the source files involved.
 *
 *  Reports on porting to new platforms, bug fixes, etc., are highly
 *  appreciated. e-mail of this kind of communication shhould be addressedd
 *  to:
 *      dvipr@i1nws1.fukui-u.ac.jp
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <math.h>
#include "newtypes.h"
#include "lips.h"
#include "gvars.h"
#include "dvi.h"
#include "font.h"
#include "tfm.h"
#include "gs.h"
#include "dvipdl.h"
#include "funcdef.h"

#define		round8(n)       (((n) + 7) >> 3)
#define		INTERNALFONT_MAGNIFICATION	(.90)

Fonts* pCurrentFont;
LipsVecParam* pLipsVecParamRoot;
static int nSplineCt;
static unsigned DWORD dwMemory = LIPS_BUILDIN_MEMORY * 1024 * 1024;
static unsigned DWORD dwMemoryMax = LIPS_BUILDIN_MEMORY * 1024 * 1024;
static double *plfSplineP = NULL, *plfSplineA = NULL, *plfSplineB = NULL, *plfSplineX = NULL, *plfSplineY = NULL;

static double GetKanjiOffset(unsigned DWORD ch);
static BOOL doLipsSpecial(char *pszSpecial);
static void uLipsMoveRelative(DWORD h, DWORD v);
static char *uLipsMakeVecParam(DWORD dwParam);
static void uLipsClearVecParam(void);
/*static char *uLipsMakeGraphicSetParam(DWORD dwParam);*/
static void uLipsCharacterSetRegistration(Fonts *pFont);
static unsigned DWORD uLipsGetCharacterSetRegistrationMemory(Fonts *pFont);
static void uLipsPolyLineStart(void);
static void uLipsOutVecParam(DWORD dwParam);
static void uLipsOutIS2(void);
static void uLipsPolygon1Start(void);
static void uLipsSetUnderlineAttribute(char *cmd);
static void uLipsSetOrnamentAttribute(char *cmd);
static void uLipsSetCharacterShadingPattern(char *cmd);
static void uLipsBeginOverlayRegistration(char *cmd);
static void uLipsOverlayOn(char *cmd);
static void uLipsOverlayOff(char *cmd);

static void LipsPutChar(WORD wCharCode);
static void LipsPutChar2(WORD wCharCode);
static void LipsReturnToText(void);
static void LipsJobStart(int nCommandLevel, WORD dwDPI, int nCharCode, char *pszJobName);
static void LipsJobEnd(void);
static void LipsColorModeDeclaration(int nMode);
static void LipsSoftReset(void);
static void LipsPaintMemoryMode(int nMode);
static void LipsSelectSizeUnit(int nMode);
static void LipsSelectPageFormat(int nMode);
static void LipsMoveAbsolute(unsigned DWORD h, unsigned DWORD v);
static void LipsMoveUp(unsigned DWORD mv);
static void LipsMoveDown(unsigned DWORD mv);
static void LipsMoveLeft(unsigned DWORD mv);
static void LipsMoveRight(unsigned DWORD mv);
static void LipsSetColor2(int nTarget, int nType, WORD nVal1, WORD nVal2, WORD nVal3);
static void LipsShiftIn(void);
static void LipsLS2(void);
static void LipsSelectCharacterPitch(WORD wPitch);
static void LipsSelectCharacterSize(WORD nSize);
static void LipsSelectCharacterSetByName(char *pszCharSetName);
static void LipsAssignNumberToCharacterSet(int nAssign, int nCharSet);
static void LipsSelectCharacterSetByNumber(int nAssign, int nTable, int nAttribute);
static void LipsCharacterSetRegistrationBin(DWORD dwTotalSize, WORD wTotalChars, WORD wGraphicSetNumber, int nDataFormat, WORD wPitch, WORD wSize, int nStyle, int nStrokeWeight, int nTypefaceNumber, int nWidth, int nHeight, int nBaseLinePos, int nUnderLinePos, int nCodeLength, int nRegisterLevel, int nBaseLineFlag, int nControlFlag, WORD wDPI, char *pData);
static void LipsAssignNameToCharacterSet(int nAssinNumber, int nSubTypeFaceNumber, WORD wStartCode, WORD wEndCode, WORD wCharacterMax, char *pszFaceName);
static void LipsCopyCharSet(int nAssignNumber, int nGraphicSetNumber, int nRegistrationLevel, char *pszCharSetName);
static void LipsSingleCharacterCopy(int nSrcCode, int nAssignNumber, int nDestCode);
static void LipsPrintControlCharacter(WORD wLength, unsigned char *pData);
static void LipsUnderlineOn(void);
static void LipsDoubleUnderlineOn(void);
static void LipsUnderlineOff(void);
static void LipsCharacterFillOn(void);
static void LipsCharacterFillOff(void);
static void LipsCharacterShadowingOn(void);
static void LipsCharacterShadowingOff(void);
static void LipsCharacterOutlineOn(void);
static void LipsCharacterOutlineOff(void);
static void LipsCharacterBoldOn(void);
static void LipsCharacterBoldOff(void);
static void LipsCharacterItalicOn(void);
static void LipsCharacterItalicOff(void);
static void LipsCharacterShadingOn(void);
static void LipsCharacterShadingOff(void);
static void LipsCharacterReverseOn(void);
static void LipsCharacterReverseOff(void);
static void LipsRasterImageBin(unsigned DWORD dwBytePerLine, unsigned DWORD h, unsigned DWORD w, int nBPP, int nDataFormat, int nOffPixel, int nImageType, unsigned char* pBits);
static void LipsEndOverlayRegistration(void);
static void LipsPaperFeedMode(int nMode);
static void LipsSetNumberOfCopies(int nPapers);
static void LipsDisplayComment2(int nLevel, char *pszMessage);
static void LipsBell(void);

static void LipsTextToVector(int nMode);
static void LipsBeginPicture(char *pszLabel);
static void LipsScalingMode(int nMode, int nMag);
static void LipsEndPicture(void);
static void LipsOriginTransfer(int x, int y);
static void LipsBeginPictureBody(void);
static void LipsVectorToText(DWORD x, DWORD y);
static void LipsLineType(int nLinePattern, int nMagnification);
static void LipsLineJoint(int nMode);
static void LipsLineWidth(int nWidth);
static void LipsInteriorStyle(int nPattern, int nRin, int nRotation, int nLeftRight, int nBlackWhite);
static void LipsRectangle(DWORD Xmin, DWORD Ymin, DWORD Xmax, DWORD Ymax);
static void LipsEllipse(int cx, int cy, int rx, int ry, int rox, int roy);
static void LipsEllipticalArc(int nType, int cx, int cy, int rx, int ry, int rox, int roy, int sx, int sy, int ex, int ey, int nArcType);

static BOOL SplineMakeTable(double x[], double y[], double z[]);
static BOOL SplineMakeTable2(TPICArray *pta);
static double Spline(double t, double x[], double y[], double z[]);
static void Spline2(double t, double *px, double *py);

BOOL LipsPdlOutChar(WORD ch)
{
	fputc(ch, gfpOut);
	return TRUE;
}

BOOL LipsPdlPrintInfo(void)
{
	DWORD	dwTotalChars, dwCharUsed;
	Fonts	*pFont;
	Gryph	*pGryph;

	LipsDisplayComment2(1, "DVIPR Printing..");
	for (pFont = pFontRoot; pFont; pFont = pFont->next) {
		dwTotalChars = dwCharUsed = 0;
		for (pGryph = pFont->pGryph; pGryph; pGryph = pGryph->next) {
			dwTotalChars++;
			dwCharUsed += pGryph->dwCounter;
		}
/*fprintf(stderr, "%s[%d]%d: total=%d, used=%d\n", pFont->acName, pFont->dwFontNum, pFont->Type, dwTotalChars, dwCharUsed);*/
		if (dwTotalChars)
			if ((double)dwCharUsed / (double)dwTotalChars > (double)gnThresholdCharacters / 100. || gfForceFontDownLoad)
				if (gfEnableFontDownLoad || gfForceFontDownLoad)
					if (dwMemory - uLipsGetCharacterSetRegistrationMemory(pFont) > (double)dwMemoryMax * (double)gnThresholdMemory / 100.) {
						uLipsCharacterSetRegistration(pFont);
						dwMemory -= uLipsGetCharacterSetRegistrationMemory(pFont);
					}
	}
	return TRUE;
}

BOOL LipsPdlSelectFont(Fonts* pFont)
{
	char	acCharSetName[64];
	static BOOL	fFlag = FALSE;

	LipsDisplayComment2(1, fFlag ? "Prepareing font" : "Wait a moment...");
	fFlag = !fFlag;
	if (!strncmp(pFont->acName, "lipsj", 5) || !strncmp(pFont->acName, "lipst", 5) || !strncmp(pFont->acName, "min", 3) || !strncmp(pFont->acName, "tmin", 4) || !strncmp(pFont->acName, "goth", 4) || !strncmp(pFont->acName, "tgoth", 5)) {
		ReadTFM(pFont);
		pFont->Type = FONT_INTERNAL;
		LipsLS2();
		if (!strncmp(pFont->acName, "lips", 4)) {
			pFont->bDirection = (!strncmp(pFont->acName, "lipst", 5) ? TRUE : FALSE);
			sprintf(acCharSetName, "%s.J78", &pFont->acName[6]);
			LipsSelectCharacterSetByName(acCharSetName);
		} else {
			pFont->bDirection = (!strncmp(pFont->acName, "tmin", 4) || !strncmp(pFont->acName, "tgoth", 5) ? TRUE : FALSE);
			if (!strncmp(pFont->acName, "min", 3) || !strncmp(pFont->acName, "tmin", 4))
				LipsSelectCharacterSetByName("Mincho-Medium.J78");
			else
				LipsSelectCharacterSetByName("Gothic-Medium.J78");
		}
		LipsSelectCharacterPitch((WORD)(7227. / ((double)pFont->dwScaled / (double)Pow(2., 16) * ((double)gdwMagnification / 1000.) * INTERNALFONT_MAGNIFICATION)));
		LipsSelectCharacterSize((WORD)((double)pFont->dwScaled / (double)Pow(2., 16) * ((double)gdwMagnification / 1000.) * INTERNALFONT_MAGNIFICATION / 72.27 * (double)LIPS_UNIT_SIZE));
		LipsAssignNumberToCharacterSet(pFont->dwFontNum, 0);
		return TRUE;
	} else if (!strncmp(pFont->acName, "lips", 4)) {
		ReadTFM(pFont);
		pFont->Type = FONT_INTERNAL_1BYTE;
		sprintf(acCharSetName, "%s.IBML", &pFont->acName[5]);
		LipsSelectCharacterSetByName(acCharSetName);
		LipsSelectCharacterSize((WORD)((double)pFont->dwScaled / (double)Pow(2., 16) * ((double)gdwMagnification / 1000.) * INTERNALFONT_MAGNIFICATION / 72.27 * (double)LIPS_UNIT_SIZE));
		LipsCopyCharSet(pFont->dwFontNum, 0xf00, 0, "");
		LipsSingleCharacterCopy(0x22, pFont->dwFontNum, 0x5c);
		LipsSingleCharacterCopy(0x22, pFont->dwFontNum, 0x7d);
		sprintf(acCharSetName, "%s.IBMR1", &pFont->acName[5]);
		LipsSelectCharacterSetByName(acCharSetName);
		LipsSelectCharacterSize((WORD)((double)pFont->dwScaled / (double)Pow(2., 16) * ((double)gdwMagnification / 1000.) * INTERNALFONT_MAGNIFICATION / 72.27 * (double)LIPS_UNIT_SIZE));
		LipsSingleCharacterCopy(0x2d, pFont->dwFontNum, 0x3c);
		LipsSingleCharacterCopy(0x28, pFont->dwFontNum, 0x3e);
		return TRUE;
	}
	return FALSE;
}

BOOL LipsPdlPrintStart(void)
{
	LipsReturnToText();
	if (gnPdlType == PDL_LIPS2)
		LipsJobStart(LIPS_JOBSTART_COMMANDLEVEL_LIPS2, gwDPI, LIPS_JOBSTART_CHARCODE_JIS, "DVIPR JOB");
	else if (gnPdlType == PDL_LIPS3)
		LipsJobStart(LIPS_JOBSTART_COMMANDLEVEL_LIPS3, gwDPI, LIPS_JOBSTART_CHARCODE_JIS, "DVIPR JOB");
	else if (gnPdlType == PDL_LIPS4)
		LipsJobStart(LIPS_JOBSTART_COMMANDLEVEL_LIPS4, gwDPI, LIPS_JOBSTART_CHARCODE_JIS, "DVIPR JOB");
	else if (gnPdlType == PDL_LIPS4C)
		LipsJobStart(LIPS_JOBSTART_COMMANDLEVEL_LIPS4C, gwDPI, LIPS_JOBSTART_CHARCODE_JIS, "DVIPR JOB");
	else if (gnPdlType == PDL_CAPSL3)
		LipsJobStart(LIPS_JOBSTART_COMMANDLEVEL_CAPSL3, gwDPI, LIPS_JOBSTART_CHARCODE_JIS, "DVIPR JOB");
	else if (gnPdlType == PDL_CAPSL4)
		LipsJobStart(LIPS_JOBSTART_COMMANDLEVEL_CAPSL4, gwDPI, LIPS_JOBSTART_CHARCODE_JIS, "DVIPR JOB");
	else {
		Error(ERR_UNKNOWN_PDL);
		FatalExit(-1);
	}
	if (gfColorMode)
		LipsColorModeDeclaration(1);
	LipsSoftReset();
	LipsSetNumberOfCopies(gnPrintPapers);
	if (gnPaperFeedMode)
		LipsPaperFeedMode(gnPaperFeedMode);
	if (gnDviVersion == 2)
		LipsSelectPageFormat(14);
	else
		LipsSelectPageFormat(15);
	if (gfColorMode)
		LipsPaintMemoryMode(0);
	fprintf(gfpOut, "%s?1l", LIPS_CSI);	/* Disable Auto NF */
	fprintf(gfpOut, "%s?2h", LIPS_CSI);	/* Disable Auto FF */
	fprintf(gfpOut, "%s?4l", LIPS_CSI);	/* Disable Auto LF at CR */
	fprintf(gfpOut, "%s?5l", LIPS_CSI);	/* Disable Auto CR at LF */
	fprintf(gfpOut, "%s?6l", LIPS_CSI);	/* Disable Auto CR at FF */
	fprintf(gfpOut, "%s?7l", LIPS_CSI);	/* Disable Auto CR at VT */
	fprintf(gfpOut, "%s?3l", LIPS_CSI);	/* Enable Auto CAP Movement */
	fprintf(gfpOut, "%s11h", LIPS_CSI); /* Size Unit Mode */
	LipsSelectSizeUnit(2);
	pLipsVecParamRoot = NULL;
	return TRUE;
}

BOOL LipsPdlPrintEnd(void)
{
	LipsDisplayComment2(1, "DVIPR done...");
	LipsJobEnd();
	return TRUE;
}

BOOL LipsPdlChangeFont(unsigned DWORD fnum)
{
	if ((pCurrentFont = GetFonts(fnum)) == NULL) {
		Error(ERR_INTERNALERROR);
		FatalExit(-1);
	}
	if (pCurrentFont->Type == FONT_INTERNAL || pCurrentFont->Type == FONT_DNPINTERNAL) {
		LipsLS2();
		LipsSelectCharacterSetByNumber(fnum, 0, 1);
		if (gnDviVersion == 3) {
			fprintf(gfpOut, "%s4 S", LIPS_CSI);
			if (pCurrentFont->bDirection)
				fprintf(gfpOut, "%s?29h", LIPS_CSI);
			else
				fprintf(gfpOut, "%s?29l", LIPS_CSI);
		}
	} else if (pCurrentFont->Type == FONT_INTERNAL_1BYTE) {
		LipsShiftIn();
		LipsSelectCharacterSetByNumber(fnum, 0, 1);
	} else {
		LipsShiftIn();
		if (xDviReg.d)
			fnum += 1024;
		if ((pCurrentFont = GetFonts(fnum)) == NULL) {
			Error(ERR_INTERNALERROR);
			FatalExit(-1);
		}
		if (pCurrentFont->Type == FONT_DOWNLOAD)
			LipsSelectCharacterSetByNumber(fnum, 1, 1);
	}
	return TRUE;
}

unsigned DWORD LipsPdlPutChar(unsigned DWORD ch)
{
	Gryph* pGryph;

	if (pCurrentFont->Type == FONT_INTERNAL || pCurrentFont->Type == FONT_DNPINTERNAL || pCurrentFont->Type == FONT_INTERNAL_1BYTE) {
		if (gnDviVersion == 2)
/*			LipsMoveAbsolute(SpToHLp(xDviReg.h - GetKanjiOffset(ch) * (DWORD)((double)GetCharWidth(pCurrentFont, 0x2121) / Pow(2., 20) * (double)pCurrentFont->dwScaled)), SpToVLp(xDviReg.v));*/
			if (pCurrentFont->Type == FONT_INTERNAL)
				LipsMoveAbsolute(SpToHLp(xDviReg.h - GetKanjiOffset(ch) * (DWORD)((double)GetCharWidth(pCurrentFont, 0x2121) / Pow(2., 20) * (double)pCurrentFont->dwScaled)), SpToVLp(xDviReg.v));
			else
				LipsMoveAbsolute(SpToHLp(xDviReg.h), SpToVLp(xDviReg.v));
		else {
			if (pCurrentFont->bDirection) {
				LipsMoveAbsolute(SpToHLp(xDviReg.v - GetKanjiOffset(ch) * (DWORD)((double)GetCharWidth(pCurrentFont, 0x2121) / Pow(2., 20) * (double)pCurrentFont->dwScaled)), (200. / 25.4 - 63. / 300.) * 720. - SpToVLp(xDviReg.h));
				uLipsMoveRelative(0, (double)pCurrentFont->dwScaled / (double)Pow(2., 16) / 72.27 * (double)LIPS_UNIT_SIZE * INTERNALFONT_MAGNIFICATION / 2.);
			} else {
				LipsMoveAbsolute(SpToHLp(xDviReg.v), (200. / 25.4 - 63. / 300.) * 720. - SpToVLp(xDviReg.h - GetKanjiOffset(ch) * (DWORD)((double)GetCharWidth(pCurrentFont, 0x2121) / Pow(2., 20) * (double)pCurrentFont->dwScaled)));
				uLipsMoveRelative(-((double)pCurrentFont->dwScaled / (double)Pow(2., 16) / 72.27 * (double)LIPS_UNIT_SIZE) * INTERNALFONT_MAGNIFICATION, 0);
			}
		}
		if (pCurrentFont->Type == FONT_INTERNAL || pCurrentFont->Type == FONT_DNPINTERNAL)
			LipsPutChar2(ch);
		else
			LipsPutChar(ch);
	} else if (pCurrentFont->Type == FONT_DOWNLOAD) {
		if ((pGryph = GetGryph(pCurrentFont, ch)) == NULL) {
			Error(ERR_INTERNALERROR);
			FatalExit(-1);
		}
		if (gnDviVersion == 2)
			LipsMoveAbsolute(SpToHLp(xDviReg.h) - JobToText(pGryph->dwHorzOffset), SpToVLp(xDviReg.v) - JobToText(pGryph->dwVertOffset));
		else {
			if (pCurrentFont->bDirection)
				LipsMoveAbsolute(SpToHLp(xDviReg.v) - JobToText(pGryph->dwVertOffset), (200. / 25.4 - 63. / 300.) * 720 - SpToVLp(xDviReg.h) - JobToText(pGryph->dwWidth - pGryph->dwHorzOffset));
			else
				LipsMoveAbsolute(SpToHLp(xDviReg.v) - JobToText(pGryph->dwHorzOffset), (200. / 25.4 - 63. / 300.) * 720 - SpToVLp(xDviReg.h) - JobToText(pGryph->dwVertOffset));
		}
		LipsPutChar(ch);
	} else {
		if ((pGryph = GetGryph(pCurrentFont, ch)) == NULL) {
			Error(ERR_INTERNALERROR);
			FatalExit(-1);
		}
		if (gnDviVersion == 2) {
			LipsMoveAbsolute(SpToHLp(xDviReg.h) - JobToText(pGryph->dwHorzOffset), SpToVLp(xDviReg.v) - JobToText(pGryph->dwVertOffset));
			LipsRasterImageBin(round8(pGryph->dwWidth), pGryph->dwHeight, pGryph->dwWidth, 1, 0, 1, 1, pGryph->pBits);
		} else {
			if (pCurrentFont->bDirection) {
				LipsMoveAbsolute(SpToHLp(xDviReg.v) - JobToText(pGryph->dwVertOffset), (200. / 25.4 - 63. / 300.) * 720 - SpToVLp(xDviReg.h) - JobToText(pGryph->dwWidth - pGryph->dwHorzOffset));
				LipsRasterImageBin(round8(pGryph->dwHeight), pGryph->dwWidth, pGryph->dwHeight, 1, 0, 1, 1, pGryph->pBits);
			} else {
				LipsMoveAbsolute(SpToHLp(xDviReg.v) - JobToText(pGryph->dwHorzOffset), (200. / 25.4 - 63. / 300.) * 720 - SpToVLp(xDviReg.h) - JobToText(pGryph->dwVertOffset));
				LipsRasterImageBin(round8(pGryph->dwWidth), pGryph->dwHeight, pGryph->dwWidth, 1, 0, 1, 1, pGryph->pBits);
			}
		}
	}
	return (DWORD)((double)GetCharWidth(pCurrentFont, ch) / Pow(2., 20) * (double)pCurrentFont->dwScaled);
}

BOOL LipsPdlPutRule(DWORD a, DWORD b)
{
	LipsMoveAbsolute(0, 0);
	LipsTextToVector(1);
	LipsBeginPicture("SetRule");
	LipsScalingMode(-1, 1);
	LipsBeginPictureBody();
fprintf(gfpOut, "}\"%s%s%s", uLipsMakeVecParam(0),  uLipsMakeVecParam((DWORD)((double)LIPS_TEXT_TOP_MARGIN * (double)LIPS_UNIT_SIZE + .5)), LIPS_IS2);
	LipsInteriorStyle(1, 1, 0, 0, 0);
	if (gnDviVersion == 2)
		LipsRectangle(SpToHLp(xDviReg.h), SpToVLp(xDviReg.v - a), SpToHLp(xDviReg.h + b), SpToVLp(xDviReg.v));
	else
		LipsRectangle(SpToHLp(xDviReg.v), (200. / 25.4 - 63. / 300.) * 720. - SpToVLp(xDviReg.h - a), SpToHLp(xDviReg.v + b), (200. / 25.4 - 63. / 300.) * 720. - SpToVLp(xDviReg.h));
	LipsEndPicture();
	LipsVectorToText(0, 0);
	return TRUE;
}

BOOL LipsPdlNewPage(WORD wPageNum)
{
#if 0
	static char *pszMessage[] = {
		">>    %3d     <<", " >>   %3d    << ", "  >>  %3d   <<  ",
		"   >> %3d  <<   ", "    >>%3d <<    ", "    Page.%3d    "
	};
	char acBuff[16 + 1];
	int	i;

	for (i = 0; i < sizeof (pszMessage) / sizeof (pszMessage[0]); i++) {
		sprintf(acBuff, pszMessage[i], wPageNum);
		LipsDisplayComment2(1, acBuff);
	}
#endif
	return TRUE;
}

BOOL LipsPdlFeedPage(void)
{
	fprintf(gfpOut, "%s", LIPS_FF);
	return TRUE;
}

BOOL LipsPdlSetColor(Color *pColor)
{
	if (pColor->Type == COLOR_CMYK)
		LipsSetColor2(0, 0, (WORD)((double)(COLOR_SCALED - pColor->dwValue[0]) * (double)pColor->dwValue[3] / COLOR_SCALED), (WORD)((double)(COLOR_SCALED - pColor->dwValue[1]) * (double)pColor->dwValue[3] / COLOR_SCALED), (WORD)((double)(COLOR_SCALED - pColor->dwValue[2]) * (double)pColor->dwValue[3] / COLOR_SCALED));
	else if (pColor->Type == COLOR_RGB)
		LipsSetColor2(0, 0, pColor->dwValue[0], pColor->dwValue[1], pColor->dwValue[2]);
	else if (pColor->Type == COLOR_GRAY)
		LipsSetColor2(0, 2, pColor->dwValue[0], pColor->dwValue[0], pColor->dwValue[0]);
	else {
		Error(ERR_COLOR_UNKNOWN_CLASS);
		return FALSE;
	}
	return TRUE;
}

BOOL LipsPdlSpecial(char *pszSpecial)
{
	BOOL	ret;
	char	*pGsDevice = NULL;

	ret = FALSE;
	if (gnPdlType == PDL_LIPS3)
		pGsDevice = "lips3fig";
	else if (gnPdlType == PDL_LIPS4)
		pGsDevice = "lips4fig";
	else if (gnPdlType == PDL_LIPS4C)
		pGsDevice = "lips4cfig";
	if (!strncmp(pszSpecial, "PSfile", 6))
		ret = SpecialPSfile(pszSpecial, pGsDevice);
	else if (!strncmp(pszSpecial, "postscriptbox", 13))
		ret = Specialpostscriptbox(pszSpecial, pGsDevice);
	else if (!strncmp(pszSpecial, "epsfile", 7))
		ret = Specialepsfile(pszSpecial, pGsDevice);
	else if (pszSpecial[0] == '\"')
		ret = SpecialRawPS(pszSpecial, pGsDevice);
	else if (!strncmp(pszSpecial, "lips:", 5))
		ret = doLipsSpecial(&pszSpecial[5]);
	return ret;
}

void LipsPdlError(ErrorCode err)
{
    char *ptr;

	LipsBell();
	ptr = SearchErrorMessage(err, ".lps");
	if (ptr != NULL)
		LipsDisplayComment2(2, ptr);
}

void LipsPdlError2(ErrorCode err, char *arg1)
{
    char	*ptr;
	char	acBuff[32];

	LipsBell();
	ptr = SearchErrorMessage(err, ".lps");
	if (ptr != NULL) {
		sprintf(acBuff, ptr, arg1);
		LipsDisplayComment2(2, acBuff);
	}
}

BOOL LipsPdlTPICDrawPolygon(DWORD dwPenSize, TPICArray *pta, BOOL bDrawFrame, BOOL bShading, double lfShading, double lfLength, double lfDistance)
{
	int nFillPattern;
	TPICArray	*ptr, tmp;

	LipsMoveAbsolute(0, 0);
	LipsTextToVector(0);
	LipsBeginPicture("TPIC");
	LipsScalingMode(-1, 1);
	LipsBeginPictureBody();
	LipsOriginTransfer(SpToHVp(xDviReg.h), SpToVVp(xDviReg.v));
	if (lfLength)
		LipsLineType(-1, 0);
	if (lfDistance)
		LipsLineType(-2, 0);
	LipsLineWidth(TPICtoVec(dwPenSize));
	LipsLineJoint(2);
	if (lfShading < .25)
		nFillPattern = 0;
	else if (lfShading < .5)
		nFillPattern = -1;
	else if (lfShading < .76)
		nFillPattern = -8;
	else
		nFillPattern = 1;
	LipsInteriorStyle(nFillPattern, bDrawFrame ? 1 : 0, 0, 0, 0);
	tmp = *pta;
	for (ptr = pta; ptr->next; ptr = ptr->next);
	if (tmp.x == ptr->x && tmp.y == ptr->y && bShading)
		uLipsPolygon1Start();
	else
		uLipsPolyLineStart();
	tmp.x = tmp.y = 0;
	for (ptr = pta; ptr; ptr = ptr->next) {
		uLipsOutVecParam(TPICtoVec(ptr->x - tmp.x));
		uLipsOutVecParam(TPICtoVec(ptr->y - tmp.y));
		tmp.x = ptr->x; tmp.y = ptr->y;
	}
	uLipsOutIS2();
	LipsEndPicture();
	LipsVectorToText(0, 0);
	return TRUE;
}

BOOL LipsPdlTPICDrawEllipse(DWORD dwPenSize, BOOL bDrawFrame, double lfShading, DWORD cx, DWORD cy, DWORD rx, DWORD ry, double srad, double erad)
{
	int nFillPattern;

	LipsMoveAbsolute(0, 0);
	LipsTextToVector(0);
	LipsBeginPicture("TPIC");
	LipsScalingMode(-1, 1);
	LipsBeginPictureBody();
	LipsOriginTransfer(SpToHVp(xDviReg.h), SpToVVp(xDviReg.v));
	LipsLineWidth(TPICtoVec(dwPenSize));
	if (lfShading < .25)
		nFillPattern = 0;
	else if (lfShading < .5)
		nFillPattern = -1;
	else if (lfShading < .76)
		nFillPattern = -8;
	else
		nFillPattern = 1;
	LipsInteriorStyle(nFillPattern, bDrawFrame ? 1 : 0, 0, 0, 0);
	if (erad - srad > 6.28)
		LipsEllipse(TPICtoVec(cx), TPICtoVec(cy), TPICtoVec(rx), TPICtoVec(ry), 1, 0);
	else
		LipsEllipticalArc(0, TPICtoVec(cx), TPICtoVec(cy), TPICtoVec(rx), TPICtoVec(ry), 1, 0, cos(srad) * 100., sin(srad) * 100., cos(erad) * 100., sin(erad) * 100., 0);
	LipsEndPicture();
	LipsVectorToText(0, 0);
	return TRUE;
}

BOOL LipsPdlTPICDrawSpline(DWORD dwPenSize, TPICArray *pta, double lfType)
{
	double	t, x, y;
	TPICArray	tmp, *end;

	LipsMoveAbsolute(0, 0);
	LipsTextToVector(0);
	LipsBeginPicture("TPIC");
	LipsScalingMode(-1, 1);
	LipsBeginPictureBody();
	LipsOriginTransfer(SpToHVp(xDviReg.h), SpToVVp(xDviReg.v));
	if (lfType > 0)
		LipsLineType(-1, 0);
	else if (lfType < 0)
		LipsLineType(-2, 0);
	LipsLineWidth(TPICtoVec(dwPenSize));
	LipsLineJoint(2);
	uLipsPolyLineStart();
	SplineMakeTable2(pta);
	for (end = pta; end->next; end = end->next);
	tmp.x = tmp.y = 0;
	for (t = 0.0; t <= 1.0; t += .005) {
		Spline2(t, &x, &y);
		uLipsOutVecParam(TPICtoVec((DWORD)x - tmp.x));
		uLipsOutVecParam(TPICtoVec((DWORD)y - tmp.y));
		tmp.x = (DWORD)x; tmp.y = (DWORD)y;
	}
	uLipsOutVecParam(TPICtoVec(end->x - tmp.x));
	uLipsOutVecParam(TPICtoVec(end->y - tmp.y));
	uLipsOutIS2();
	LipsEndPicture();
	LipsVectorToText(0, 0);
	return TRUE;
}

static double GetKanjiOffset(unsigned DWORD ch)
{
	int		i;
	static KanjiOffset OffsetData[] = {
		{0x2136, 0.12}, {0x2137, 0.12}, {0x2139, 0.12},
		{0x2168, 0.12}, {0x2169, 0.12}, {0x216a, 0.12},
		{0x2170, 0.12}, {0x2171, 0.12}, {0x2172, 0.12},
		{0x2421, 0.12}, {0x2423, 0.12}, {0x2445, 0.12},
		{0x2427, 0.12}, {0x2429, 0.12}, {0x2443, 0.12},
		{0x2463, 0.12}, {0x2465, 0.12}, {0x2467, 0.12},
		{0x2521, 0.12}, {0x2523, 0.12}, {0x2545, 0.12},
		{0x2527, 0.12}, {0x2529, 0.12}, {0x2543, 0.12},
		{0x2563, 0.12}, {0x2565, 0.12}, {0x2567, 0.12},
		{0x2575, 0.12}, {0x2576, 0.12},

		{0x2126, 0.33}, {0x2127, 0.33}, {0x2128, 0.33},
		{0x212a, 0.33}, {0x212d, 0.33}, {0x212e, 0.33},
		{0x213e, 0.33}, {0x2142, 0.33}, {0x2143, 0.33},
		{0x2146, 0.65},

		{0x2129, 0.25}, {0x212f, 0.25}, {0x2130, 0.25},
		{0x2133, 0.25}, {0x2135, 0.25}, {0x2148, 0.50},
		{0x2178, 0.25},

		{0x214a, 0.50}, {0x214c, 0.50}, {0x214e, 0.50},
		{0x2150, 0.50},	{0x2152, 0.50}, {0x2154, 0.50},
		{0x2156, 0.50},	{0x2158, 0.50}, {0x215a, 0.50},

		{0x2426, 0.02},	{0x242f, 0.02}, {0x2430, 0.02},
		{0x2431, 0.02},	{0x2432, 0.02}, {0x2439, 0.02},
		{0x243a, 0.02},	{0x246a, 0.02}, {0x2546, 0.02},
		{0x2547, 0.02},	{0x2564, 0.02},

		{0x2522, 0.02},	{0x2524, 0.02}, {0x2526, 0.02},
		{0x252a, 0.02},	{0x2531, 0.02}, {0x2532, 0.02},
		{0x2535, 0.02},	{0x2536, 0.02}, {0x253d, 0.02},
		{0x253e, 0.02},	{0x2541, 0.02}, {0x2542, 0.02},
		{0x2544, 0.02},	{0x2545, 0.02}, {0x2548, 0.02},
		{0x2549, 0.02},	{0x254a, 0.02}, {0x2555, 0.02},
		{0x2556, 0.02},	{0x2557, 0.02}, {0x255f, 0.02},
		{0x2561, 0.02},	{0x2569, 0.02}, {0x256a, 0.02},
		{0x256f, 0.02},	{0x2572, 0.02},

		{0x252f, 0.02},	{0x2530, 0.02}, {0x253f, 0.02},
		{0x2540, 0.02},

		{0x254e, 0.02},
	};

	for (i = 0; i < sizeof (OffsetData) / sizeof (OffsetData[0]); i++)
		if (ch == OffsetData[i].ch)
			return OffsetData[i].fOffset;
	return 0.;
}

static BOOL doLipsSpecial(char *pszSpecial)
{
	LipsMoveAbsolute(SpToHLp(xDviReg.h), SpToVLp(xDviReg.v));
	if (!strncmp(pszSpecial, "SETUDA", 6))
		uLipsSetUnderlineAttribute(&pszSpecial[7]);
	else if (!strncmp(pszSpecial, "BEGUDL", 6))
		LipsUnderlineOn();
	else if (!strncmp(pszSpecial, "BEGDUL", 6))
		LipsDoubleUnderlineOn();
	else if (!strncmp(pszSpecial, "ENDUDL", 6))
		LipsUnderlineOff();
	else if (!strncmp(pszSpecial, "SETOMA", 6))
		uLipsSetOrnamentAttribute(&pszSpecial[7]);
	else if (!strncmp(pszSpecial, "BEGCHF", 6))
		LipsCharacterFillOn();
	else if (!strncmp(pszSpecial, "ENDCHF", 6))
		LipsCharacterFillOff();
	else if (!strncmp(pszSpecial, "BEGCHS", 6))
		LipsCharacterShadowingOn();
	else if (!strncmp(pszSpecial, "ENDCHS", 6))
		LipsCharacterShadowingOff();
	else if (!strncmp(pszSpecial, "BEGCHO", 6))
		LipsCharacterOutlineOn();
	else if (!strncmp(pszSpecial, "ENDCHO", 6))
		LipsCharacterOutlineOff();
	else if (!strncmp(pszSpecial, "BEGCHB", 6))
		LipsCharacterBoldOn();
	else if (!strncmp(pszSpecial, "ENDCHB", 6))
		LipsCharacterBoldOff();
	else if (!strncmp(pszSpecial, "BEGCHI", 6))
		LipsCharacterItalicOn();
	else if (!strncmp(pszSpecial, "ENDCHI", 6))
		LipsCharacterItalicOff();
	else if (!strncmp(pszSpecial, "SETCSH", 6))
		uLipsSetCharacterShadingPattern(&pszSpecial[7]);
	else if (!strncmp(pszSpecial, "BEGCSH", 6))
		LipsCharacterShadingOn();
	else if (!strncmp(pszSpecial, "ENDCSH", 6))
		LipsCharacterShadingOff();
	else if (!strncmp(pszSpecial, "BEGCRV", 6))
		LipsCharacterReverseOn();
	else if (!strncmp(pszSpecial, "ENDCRV", 6))
		LipsCharacterReverseOff();
	else if (!strncmp(pszSpecial, "BEGOVR", 6))
		uLipsBeginOverlayRegistration(&pszSpecial[7]);
	else if (!strncmp(pszSpecial, "ENDOVR", 6))
		LipsEndOverlayRegistration();
	else if (!strncmp(pszSpecial, "BEGOVL", 6))
		uLipsOverlayOn(&pszSpecial[7]);
	else if (!strncmp(pszSpecial, "ENDOVL", 6))
		uLipsOverlayOff(&pszSpecial[7]);
	else
		return FALSE;
	return TRUE;
}

/**************************/
/***   USER  FUNCTION   ***/
/**************************/

static void uLipsMoveRelative(DWORD h, DWORD v)
{
	if (h)
		if (h < 0)
			LipsMoveLeft(-h);
		else
			LipsMoveRight(h);
	if (v)
		if (v < 0)
			LipsMoveUp(-v);
		else
			LipsMoveDown(v);
}

static char *uLipsMakeVecParam(DWORD dwParam)
{
	LipsVecParam *ptr, *pNewLipsVecParam;
	int	i, ret;
	BOOL bSign;

	if ((pNewLipsVecParam = (LipsVecParam*)malloc(sizeof (LipsVecParam))) == NULL) {
		Error(ERR_NOMEMORY);
		FatalExit(-1);
	}
	memset(pNewLipsVecParam, 0, sizeof (LipsVecParam));
	if (pLipsVecParamRoot == NULL)
		pLipsVecParamRoot = pNewLipsVecParam;
	else {
		for (ptr = pLipsVecParamRoot; ptr->next; ptr = ptr->next);
		ptr->next = pNewLipsVecParam;
	}
	bSign = TRUE;
	if (dwParam < 0) {
		bSign = FALSE;
		dwParam = -dwParam;
	}
	pNewLipsVecParam->acParam[5] = (dwParam & 0x0f) | 0x20 | (bSign ? 0x10 : 0x00);
	dwParam >>= 4;
	for (i = ret = 5; i; i--) {
		if (dwParam)
			ret = i - 1;
		pNewLipsVecParam->acParam[i - 1] = (dwParam & 0x3f) | 0x40;
		dwParam >>= 6;
	}
	return &(pNewLipsVecParam->acParam[ret]);
}

static void uLipsClearVecParam(void)
{
	LipsVecParam	*ptr, *next;

	for (ptr = pLipsVecParamRoot; ptr; ptr = next) {
		next = ptr->next;
		free(ptr);
	}
	pLipsVecParamRoot = NULL;
}
/*
static char *uLipsMakeGraphicSetParam(DWORD dwParam)
{
	int			i;
	static char	achBuff[8 + 1];

	memset(achBuff, 0, 8 + 1);
	for (i = 8; i; i--) {
		achBuff[i - 1] = (dwParam & 0x0f) | (i == 8 ? 0x30 : 0x20);
		dwParam >>= 4;
	}
	for (i = 0; i < 8; i++)
		if (achBuff[i] != ' ')
			break;
	return &(achBuff[i]);
}
*/
static void uLipsCharacterSetRegistration(Fonts *pFont)
{
	unsigned DWORD	dwWidthMax, dwHeightMax, dwGryphDataSize, dwHeight, dwWidth;
	char			*pGryphData, *ptr, acMessage[32], acFaceName[64];
	Gryph			*pGryph;
	WORD			wPitch, wSize;

	dwWidthMax = GetFontWidthMax(pFont) + 1;
	dwHeightMax = GetFontHeightMax(pFont) + 1;
	if (pFont->bDirection)
		dwGryphDataSize = (1 + round8(dwHeightMax) * dwWidthMax) * GetFontGryphs(pFont);
	else
		dwGryphDataSize = (1 + round8(dwWidthMax) * dwHeightMax) * GetFontGryphs(pFont);
	if ((pGryphData = (char *)malloc(dwGryphDataSize)) == NULL) {
		Error(ERR_NOMEMORY);
		FatalExit(-1);
	}
	memset(pGryphData, 0, dwGryphDataSize);
	ptr = pGryphData;
	for (pGryph = pFont->pGryph; pGryph != NULL; pGryph = pGryph->next) {
		*((unsigned char *)ptr) = pGryph->dwCharCode;
		ptr += sizeof (unsigned char);
		if (pFont->bDirection) {
			for (dwWidth = 0; dwWidth < pGryph->dwWidth; dwWidth++)
				memcpy(&ptr[round8(dwHeightMax) * dwWidth], &pGryph->pBits[round8(pGryph->dwHeight) * dwWidth], round8(pGryph->dwHeight));
			ptr += round8(dwHeightMax) * dwWidthMax;
		} else {
			for (dwHeight = 0; dwHeight < pGryph->dwHeight; dwHeight++)
				memcpy(&ptr[round8(dwWidthMax) * dwHeight], &pGryph->pBits[round8(pGryph->dwWidth) * dwHeight], round8(pGryph->dwWidth));
			ptr += round8(dwWidthMax) * dwHeightMax;
		}
	}
	sprintf(acMessage, "Loading %8.8s", pFont->acName);
	LipsDisplayComment2(1, acMessage);
	sprintf(acFaceName, "%.40s.%7.7s", pFont->acName, pFont->acName);
	LipsAssignNameToCharacterSet(pFont->dwFontNum, 10, 0, 0, GetFontGryphs(pFont), acFaceName);
	wPitch = (WORD)(7227. / ((double)pFont->dwScaled / (double)Pow(2., 16) * ((double)gdwMagnification / 1000.)));
	wSize  = (WORD)((double)pFont->dwScaled / (double)Pow(2., 16) * ((double)gdwMagnification / 1000.) * 100.);
	if (pFont->bDirection)
		LipsCharacterSetRegistrationBin(dwGryphDataSize, GetFontGryphs(pFont), 0xf00, 0, wPitch, wSize, 0, 0, 200, dwHeightMax, dwWidthMax, dwWidthMax, 0, 0, 0, 0, 1, gwDPI, pGryphData);
	else
		LipsCharacterSetRegistrationBin(dwGryphDataSize, GetFontGryphs(pFont), 0xf00, 0, wPitch, wSize, 0, 0, 200, dwWidthMax, dwHeightMax, dwHeightMax, 0, 0, 0, 0, 1, gwDPI, pGryphData);
	pFont->Type = FONT_DOWNLOAD;
	free(pGryphData);
}

static unsigned DWORD uLipsGetCharacterSetRegistrationMemory(Fonts *pFont)
{
	unsigned DWORD	dwUsedMemory;

	dwUsedMemory = 84;		/* Information header about character set */
	dwUsedMemory += GetFontGryphs(pFont) * 4;	/* Pointer to characters */
	dwUsedMemory += 52;		/* Name of character set */
	dwUsedMemory += (round8(GetFontWidthMax(pFont)) * GetFontHeightMax(pFont) + 20) * GetFontGryphs(pFont); /* Part of the pattern */
	dwUsedMemory += 20;		/* Header to manage it */
	dwUsedMemory = (dwUsedMemory + 0xff) & ~0xff; /* Must be 256 byte boundary */
	return dwUsedMemory;
}

static void uLipsPolyLineStart(void)
{
	fprintf(gfpOut, "1");
}

static void uLipsOutVecParam(DWORD dwParam)
{
	fprintf(gfpOut, "%s", uLipsMakeVecParam(dwParam));
	uLipsClearVecParam();
}

static void uLipsOutIS2(void)
{
	fprintf(gfpOut, "%s", LIPS_IS2);
}

static void uLipsPolygon1Start(void)
{
	fprintf(gfpOut, "2");
}

static void uLipsSetUnderlineAttribute(char *cmd)
{
	fprintf(gfpOut, "%s%s%%u", LIPS_CSI, cmd);
}

static void uLipsSetOrnamentAttribute(char *cmd)
{
	fprintf(gfpOut, "%s%s%%}", LIPS_CSI, cmd);
}

static void uLipsSetCharacterShadingPattern(char *cmd)
{
	fprintf(gfpOut, "%s%s$s", LIPS_CSI, cmd);
}

static void uLipsBeginOverlayRegistration(char *cmd)
{
	fprintf(gfpOut, "%s1;%s;0&s", LIPS_CSI, cmd);
}

static void uLipsOverlayOn(char *cmd)
{
	fprintf(gfpOut, "%s1;%s&w", LIPS_CSI, cmd);
}

static void uLipsOverlayOff(char *cmd)
{
	fprintf(gfpOut, "%s0;%s&w", LIPS_CSI, cmd);
}

/**************************/
/*** TEXT MODE COMMANDS ***/
/**************************/

static void LipsPutChar(WORD wCharCode)
{
	unsigned char	ch;
	
	if (wCharCode == 0x00 || (0x07 <= wCharCode && wCharCode <= 0x0f) || wCharCode == 0x1b) {
		ch = wCharCode;
		LipsPrintControlCharacter(1, &ch);
	} else
		fputc(wCharCode, gfpOut);
}

static void LipsPutChar2(WORD wCharCode)
{
	fprintf(gfpOut, "%c%c", (wCharCode >> 8) & 0xff, wCharCode & 0xff);
}

static void LipsReturnToText(void)
{
	fprintf(gfpOut, "%s%%@", LIPS_ESC);
}

static void LipsJobStart(int nCommandLevel, WORD dwDPI, int nCharCode, char *pszJobName)
{
	fprintf(gfpOut, "%s%d;%d;%dJ%s%s", LIPS_DCS, nCommandLevel, dwDPI, nCharCode, pszJobName, LIPS_ST);
}

static void LipsJobEnd(void)
{
	fprintf(gfpOut, "%s0J%s", LIPS_DCS, LIPS_ST);
}

static void LipsColorModeDeclaration(int nMode)
{
	fprintf(gfpOut, "%s%d\"p", LIPS_CSI, nMode);
}

static void LipsSoftReset(void)
{
	fprintf(gfpOut, "%s<", LIPS_ESC);
}

static void LipsPaintMemoryMode(int nMode)
{
	fprintf(gfpOut, "%s%d&z", LIPS_CSI, nMode);
}

static void LipsSelectSizeUnit(int nMode)
{
	fprintf(gfpOut, "%s%d I", LIPS_CSI, nMode);
}

static void LipsSelectPageFormat(int nMode)
{
	fprintf(gfpOut, "%s%dp", LIPS_CSI, nMode);
}

static void LipsMoveAbsolute(unsigned DWORD h, unsigned DWORD v)
{
	fprintf(gfpOut, "%s%d;%df", LIPS_CSI, v, h);
}

static void LipsMoveUp(unsigned DWORD mv)
{
	fprintf(gfpOut, "%s%dk", LIPS_CSI, mv);
}

static void LipsMoveDown(unsigned DWORD mv)
{
	fprintf(gfpOut, "%s%de", LIPS_CSI, mv);
}

static void LipsMoveLeft(unsigned DWORD mv)
{
	fprintf(gfpOut, "%s%dj", LIPS_CSI, mv);
}

static void LipsMoveRight(unsigned DWORD mv)
{
	fprintf(gfpOut, "%s%da", LIPS_CSI, mv);
}

static void LipsSetColor2(int nTarget, int nType, WORD nVal1, WORD nVal2, WORD nVal3)
{
	fprintf(gfpOut, "%s?%d;%d;%d;%d;%d;%%p", LIPS_CSI, nTarget, nType, nVal1, nVal2, nVal3);
}

static void LipsShiftIn(void)
{
	fprintf(gfpOut, "%s", LIPS_SI);
}

static void LipsLS2(void)
{
	fprintf(gfpOut, "%sn", LIPS_ESC);
}

static void LipsSelectCharacterPitch(WORD wPitch)
{
	fprintf(gfpOut, "%s?%d K", LIPS_CSI, wPitch);
}

static void LipsSelectCharacterSize(WORD nSize)
{
	fprintf(gfpOut, "%s%d C", LIPS_CSI, nSize);
}

static void LipsSelectCharacterSetByName(char *pszCharSetName)
{
	fprintf(gfpOut, "%sz%s%s\\", LIPS_DCS, pszCharSetName, LIPS_ESC);
}

static void LipsAssignNumberToCharacterSet(int nAssign, int nCharSet)
{
	fprintf(gfpOut, "%s%d;%d D", LIPS_CSI, nAssign, nCharSet);
}

static void LipsSelectCharacterSetByNumber(int nAssign, int nTable, int nAttribute)
{
	fprintf(gfpOut, "%s%d;%d;%d%%v", LIPS_CSI, nAssign, nTable, nAttribute);
}

static void LipsCharacterSetRegistrationBin(DWORD dwTotalSize, WORD wTotalChars, WORD wGraphicSetNumber, int nDataFormat, WORD wPitch, WORD wSize, int nStyle, int nStrokeWeight, int nTypefaceNumber, int nWidth, int nHeight, int nBaseLinePos, int nUnderLinePos, int nCodeLength, int nRegisterLevel, int nBaseLineFlag, int nControlFlag, WORD wDPI, char *pData)
{
	fprintf(gfpOut, "%s%d;%d;0;0;%d;%d;%d;%d;%d;%d;%d;%d;%d;%d;%d;%d;%d;%d;%d;%d.p", LIPS_CSI, dwTotalSize, wTotalChars, wGraphicSetNumber, nDataFormat, wPitch, wSize, nStyle, nStrokeWeight, nTypefaceNumber, nWidth, nHeight, nBaseLinePos, nUnderLinePos, nCodeLength, nRegisterLevel, nBaseLineFlag, nControlFlag, wDPI);
	fwrite(pData, dwTotalSize, 1, gfpOut);
}

static void LipsAssignNameToCharacterSet(int nAssignNumber, int nSubTypeFaceNumber, WORD wStartCode, WORD wEndCode, WORD wCharacterMax, char *pszFaceName)
{
	fprintf(gfpOut, "%s%d;0;%d;%d;%d;%dx%s%s\\", LIPS_DCS, nAssignNumber, nSubTypeFaceNumber, wStartCode, wEndCode, wCharacterMax, pszFaceName, LIPS_ESC);
}

static void LipsCopyCharSet(int nAssignNumber, int nGraphicSetNumber, int nRegistrationLevel, char *pszCharSetName)
{
	fprintf(gfpOut, "%s%d;%d;%dp%s%s\\", LIPS_DCS, nAssignNumber, nGraphicSetNumber, nRegistrationLevel, pszCharSetName, LIPS_ESC);
}

static void LipsSingleCharacterCopy(int nSrcCode, int nAssignNumber, int nDestCode)
{
	fprintf(gfpOut, "%s%d;%d;%d&q", LIPS_CSI, nSrcCode, nAssignNumber, nDestCode);
}

static void LipsPrintControlCharacter(WORD wLength, unsigned char *pData)
{
	fprintf(gfpOut, "%s%d.v", LIPS_CSI, wLength);
	fwrite(pData, wLength, 1, gfpOut);
}

static void LipsUnderlineOn(void)
{
	fprintf(gfpOut, "%s4m", LIPS_CSI);
}

static void LipsDoubleUnderlineOn(void)
{
	fprintf(gfpOut, "%s21m", LIPS_CSI);
}

static void LipsUnderlineOff(void)
{
	fprintf(gfpOut, "%s24m", LIPS_CSI);
}

static void LipsCharacterFillOn(void)
{
	fprintf(gfpOut, "%s?5m", LIPS_CSI);
}

static void LipsCharacterFillOff(void)
{
	fprintf(gfpOut, "%s?25m", LIPS_CSI);
}

static void LipsCharacterShadowingOn(void)
{
	fprintf(gfpOut, "%s?6m", LIPS_CSI);
}

static void LipsCharacterShadowingOff(void)
{
	fprintf(gfpOut, "%s?26m", LIPS_CSI);
}

static void LipsCharacterOutlineOn(void)
{
	fprintf(gfpOut, "%s?7m", LIPS_CSI);
}

static void LipsCharacterOutlineOff(void)
{
	fprintf(gfpOut, "%s?27m", LIPS_CSI);
}

static void LipsCharacterBoldOn(void)
{
	fprintf(gfpOut, "%s?8m", LIPS_CSI);
}

static void LipsCharacterBoldOff(void)
{
	fprintf(gfpOut, "%s?28m", LIPS_CSI);
}

static void LipsCharacterItalicOn(void)
{
	fprintf(gfpOut, "%s?9m", LIPS_CSI);
}

static void LipsCharacterItalicOff(void)
{
	fprintf(gfpOut, "%s?29m", LIPS_CSI);
}

static void LipsCharacterShadingOn(void)
{
	fprintf(gfpOut, "%s5m", LIPS_CSI);
}

static void LipsCharacterShadingOff(void)
{
	fprintf(gfpOut, "%s25m", LIPS_CSI);
}

static void LipsCharacterReverseOn(void)
{
	fprintf(gfpOut, "%s7m", LIPS_CSI);
}

static void LipsCharacterReverseOff(void)
{
	fprintf(gfpOut, "%s27m", LIPS_CSI);
}

static void LipsRasterImageBin(unsigned DWORD dwBytePerLine, unsigned DWORD h, unsigned DWORD w, int nBPP, int nDataFormat, int nOffPixel, int nImageType, unsigned char* pBits)
{
	unsigned DWORD	ct;

	if (gnPdlType == PDL_LIPS3)
		fprintf(gfpOut, "%s%d;%d;%d;%d;%d.r", LIPS_CSI, dwBytePerLine * h, dwBytePerLine, gwDPI, 0, h);
	else
		fprintf(gfpOut, "%s%d;%d;%d;%d;%d;%d;%d;%d;%d;%d.r", LIPS_CSI, dwBytePerLine * h, dwBytePerLine, gwDPI, 0, h, w, nBPP, nDataFormat, nOffPixel, nImageType);
	for (ct = round8(w) * h; ct; ct--)
		fputc(*pBits++, gfpOut);
}

static void LipsEndOverlayRegistration(void)
{
	fprintf(gfpOut, "%s0&s", LIPS_CSI);
	LipsPdlChangeFont(pCurrentFont->dwFontNum);
}

static void LipsPaperFeedMode(int nMode)
{
	if (gnPdlType == PDL_LIPS3 || gnPdlType == PDL_CAPSL3 || !nMode)
		fprintf(gfpOut, "%s%dq", LIPS_CSI, nMode);
	else
		fprintf(gfpOut, "%s%dq", LIPS_CSI, nMode + 9);
}

static void LipsSetNumberOfCopies(int nPapers)
{
	fprintf(gfpOut, "%s%dv", LIPS_CSI, nPapers);
}

static void LipsDisplayComment2(int nLevel, char *pszMessage)
{
	fprintf(gfpOut, "%s%dy%s%s", LIPS_DCS, nLevel, pszMessage, LIPS_ST);
	fflush(gfpOut);
}

static void LipsBell(void)
{
	fprintf(gfpOut, "%s", LIPS_BEL);
}

/****************************/
/*** VECTOR MODE COMMANDS ***/
/****************************/

static void LipsTextToVector(int nMode)
{
	fprintf(gfpOut, "%s%d&}", LIPS_CSI, nMode);
}

static void LipsBeginPicture(char *pszLabel)
{
	fprintf(gfpOut, "#%s%s", pszLabel, LIPS_IS2);
}

static void LipsScalingMode(int nMode, int nMag)
{
	fprintf(gfpOut, "!0%s%s%s", uLipsMakeVecParam(nMode), uLipsMakeVecParam(nMag), LIPS_IS2);
	uLipsClearVecParam();
}

static void LipsBeginPictureBody(void)
{
	fprintf(gfpOut, "$%s", LIPS_IS2);
}

static void LipsEndPicture(void)
{
	fprintf(gfpOut, "%%%s", LIPS_IS2);
}

static void LipsOriginTransfer(int x, int y)
{
	fprintf(gfpOut, "}\"%s%s%s", uLipsMakeVecParam(x), uLipsMakeVecParam(y), LIPS_IS2);
	uLipsClearVecParam();
}


static void LipsVectorToText(DWORD x, DWORD y)
{
	fprintf(gfpOut, "}p%s", LIPS_IS2);
}

static void LipsLineType(int nLinePattern, int nMagnification)
{
	fprintf(gfpOut, "E1%s%s%s", uLipsMakeVecParam(nLinePattern), uLipsMakeVecParam(nMagnification), LIPS_IS2);
	uLipsClearVecParam();
}

static void LipsLineJoint(int nMode)
{
	fprintf(gfpOut, "}F%s%s", uLipsMakeVecParam(nMode), LIPS_IS2);
	uLipsClearVecParam();
}

static void LipsLineWidth(int nWidth)
{
	fprintf(gfpOut, "F1%s%s", uLipsMakeVecParam(nWidth), LIPS_IS2);
	uLipsClearVecParam();
}

static void LipsInteriorStyle(int nPattern, int nRin, int nRotation, int nLeftRight, int nBlackWhite)
{
	fprintf(gfpOut, "I%s%s%s%s%s%s", uLipsMakeVecParam(nPattern), uLipsMakeVecParam(nRin), uLipsMakeVecParam(nRotation), uLipsMakeVecParam(nLeftRight), uLipsMakeVecParam(nBlackWhite), LIPS_IS2);
	uLipsClearVecParam();
}

static void LipsRectangle(DWORD Xmin, DWORD Ymin, DWORD Xmax, DWORD Ymax)
{
	fprintf(gfpOut, "}:%s%s%s%s%s", uLipsMakeVecParam(Xmin), uLipsMakeVecParam(Xmax), uLipsMakeVecParam(Ymin), uLipsMakeVecParam(Ymax), LIPS_IS2);
	uLipsClearVecParam();
}

static void LipsEllipse(int cx, int cy, int rx, int ry, int rox, int roy)
{
	fprintf(gfpOut, "}7%s%s%s%s%s%s%s", uLipsMakeVecParam(cx), uLipsMakeVecParam(cy), uLipsMakeVecParam(rx), uLipsMakeVecParam(ry), uLipsMakeVecParam(rox), uLipsMakeVecParam(roy), LIPS_IS2);
	uLipsClearVecParam();
}

static void LipsEllipticalArc(int nType, int cx, int cy, int rx, int ry, int rox, int roy, int sx, int sy, int ex, int ey, int nArcType)
{
	fprintf(gfpOut, "};%s%s%s%s%s%s%s%s%s%s%s%s%s", uLipsMakeVecParam(nType), uLipsMakeVecParam(cx), uLipsMakeVecParam(cy), uLipsMakeVecParam(rx), uLipsMakeVecParam(ry), uLipsMakeVecParam(rox), uLipsMakeVecParam(roy), uLipsMakeVecParam(sx), uLipsMakeVecParam(sy), uLipsMakeVecParam(ex), uLipsMakeVecParam(ey), uLipsMakeVecParam(nArcType), LIPS_IS2);
	uLipsClearVecParam();
}

static BOOL SplineMakeTable(double x[], double y[], double z[])
{
	int i;
	double t;
	double *plfSplineH, *plfSplineD;

	if ((plfSplineH = (double *)malloc(sizeof (double) * nSplineCt)) == NULL || (plfSplineD = (double *)malloc(sizeof (double) * nSplineCt)) == NULL) {
		Error(ERR_NOMEMORY);
		return FALSE;
	}
	z[0] = 0; z[nSplineCt - 1] = 0;
	for (i = 0; i < nSplineCt - 1; i++) {
		plfSplineH[i    ] =  x[i + 1] - x[i];
		plfSplineD[i + 1] = (y[i + 1] - y[i]) / plfSplineH[i];
	}
	z[1] = plfSplineD[2] - plfSplineD[1] - plfSplineH[0] * z[0];
	plfSplineD[1] = 2 * (x[2] - x[0]);
	for (i = 1; i < nSplineCt - 2; i++) {
		t = plfSplineH[i] / plfSplineD[i];
		z[i + 1] = plfSplineD[i + 2] - plfSplineD[i + 1] - z[i] * t;
		plfSplineD[i + 1] = 2 * (x[i + 2] - x[i]) - plfSplineH[i] * t;
	}
	z[nSplineCt - 2] -= plfSplineH[nSplineCt - 2] * z[nSplineCt - 1];
	for (i = nSplineCt - 2; i > 0; i--)
		z[i] = (z[i] - plfSplineH[i] * z[i + 1]) / plfSplineD[i];
	free(plfSplineD);
	free(plfSplineH);
	return TRUE;
}

static double Spline(double t, double x[], double y[], double z[])
{
	int i, j, k;
	double d, h;

	i = 0; j = nSplineCt - 1;
	while (i < j) {
		k = (i + j) / 2;
		if (x[k] < t)
			i = k + 1;
		else
			j = k;
	}
	if (i > 0)
		i--;
	h = x[i + 1] - x[i];
	d = t - x[i];
	return (((z[i + 1] - z[i]) * d / h + z[i] * 3) * d + ((y[i + 1] - y[i]) / h - (z[i] * 2 + z[i + 1]) * h)) * d + y[i];
}

static BOOL SplineMakeTable2(TPICArray *pta)
{
	int i;
	double t1, t2;
	TPICArray* ptr;
	
	if (plfSplineP != NULL) {
		free(plfSplineP);
		free(plfSplineA);
		free(plfSplineB);
		free(plfSplineX);
		free(plfSplineY);
	}
	nSplineCt = 0;
	for (ptr = pta; ptr; ptr = ptr->next)
		nSplineCt++;
	if ((plfSplineP = (double *)malloc(sizeof (double) * nSplineCt)) == NULL || (plfSplineA = (double *)malloc(sizeof (double) * nSplineCt)) == NULL || (plfSplineB = (double *)malloc(sizeof (double) * nSplineCt)) == NULL || (plfSplineX = (double *)malloc(sizeof (double) * nSplineCt)) == NULL || (plfSplineY = (double *)malloc(sizeof (double) * nSplineCt)) == NULL) {
		Error(ERR_NOMEMORY);
		return FALSE;
	}
	i = 0;
	for (ptr = pta; ptr; ptr = ptr->next) {
		plfSplineX[i] = ptr->x;
		plfSplineY[i] = ptr->y;
		i++;
	}
	plfSplineP[0] = 0;
	for (i = 1; i < nSplineCt; i++) {
		t1 = plfSplineX[i] - plfSplineX[i - 1];
		t2 = plfSplineY[i] - plfSplineY[i - 1];
		plfSplineP[i] = plfSplineP[i - 1] + sqrt(t1 * t1 + t2 * t2);
	}
	for (i = 1; i < nSplineCt; i++)
		plfSplineP[i] /= plfSplineP[nSplineCt - 1];
	if (!SplineMakeTable(plfSplineP, plfSplineX, plfSplineA))
		return FALSE;
	return SplineMakeTable(plfSplineP, plfSplineY, plfSplineB);
}

static void Spline2(double t, double *px, double *py)
{
	*px = Spline(t, plfSplineP, plfSplineX, plfSplineA);
	*py = Spline(t, plfSplineP, plfSplineY, plfSplineB);
}
