/*
 *  dvipr
 *      TeX Printer driver for The Canon LIPS III/IVc, CaPSL III printer
 *
 *      Developed at:
 *          Computer Systems Research Group
 *          Department of Information Science
 *          Fukui University
 *          9-1, Bunkyo-3, Fukui, 910 Japan
 *
 *      Authors:
 *          Yasumasa Koizumi, Atsushi Takagi, Kenichi Sugimoto,
 *          and Minato Kawaguti
 *
 *  The program may be copied and redistributed freely in the spirit of
 *  Free Software Foundation's "Copy Left", provided this title page is
 *  included in all of the source files involved.
 *
 *  Reports on porting to new platforms, bug fixes, etc., are highly
 *  appreciated. e-mail of this kind of communication shhould be addressedd
 *  to:
 *      dvipr@i1nws1.fukui-u.ac.jp
 */

#ifndef __LIPS_H
#define	__LIPS_H

#include "gvars.h"
#include "font.h"
#include "color.h"
#include "tpic.h"
#include "error.h"

/*** C0 Control Code ***/
#define	LIPS_NUL	"\x00"
#define	LIPS_EXT	"\x03"
#define	LIPS_ACK	"\x06"
#define	LIPS_BEL	"\x07"
#define	LIPS_BS		"\x08"
#define	LIPS_HT		"\x09"
#define	LIPS_LF		"\x0a"
#define	LIPS_VT		"\x0b"
#define	LIPS_FF		"\x0c"
#define	LIPS_CR		"\x0d"
#define	LIPS_SO		"\x0e"
#define	LIPS_SI		"\x0f"
#define	LIPS_XON	"\x11"
#define	LIPS_XOFF	"\x13"
#define	LIPS_EM		"\x19"
#define	LIPS_ESC	"\x1b"
#define	LIPS_GS		"\x1d"
#define	LIPS_IS2	"\x1e"

/*** C0 Control Code ***/
#ifdef DVIPR_8BIT
#define	LIPS_NEL	"\x85"
#define	LIPS_HTS	"\x88"
#define	LIPS_VTS	"\x8a"
#define	LIPS_PLD	"\x8b"
#define	LIPS_PLU	"\x8c"
#define	LIPS_RI		"\x8d"
#define	LIPS_SS2	"\x8e"
#define	LIPS_SS3	"\x8f"
#define	LIPS_DCS	"\x90"
#define	LIPS_CSI	"\x9b"
#define	LIPS_ST		"\x9c"
#define	LIPS_SP		"\x20"
#define	LIPS_DEL	"\x7f"
#else
#define	LIPS_NEL	LIPS_ESC "E"
#define	LIPS_HTS	LIPS_ESC "H"
#define	LIPS_VTS	LIPS_ESC "J"
#define	LIPS_PLD	LIPS_ESC "K"
#define	LIPS_PLU	LIPS_ESC "L"
#define	LIPS_RI		LIPS_ESC "M"
#define	LIPS_SS2	LIPS_ESC "N"
#define	LIPS_SS3	LIPS_ESC "O"
#define	LIPS_DCS	LIPS_ESC "P"
#define	LIPS_CSI	LIPS_ESC "["
#define	LIPS_ST		LIPS_ESC "\\"
#define	LIPS_SP		"\x20"
#define	LIPS_DEL	"\x7f"
#endif

/*** Command Level ***/
#define	LIPS_JOBSTART_COMMANDLEVEL_LIPS2		21
#define	LIPS_JOBSTART_COMMANDLEVEL_LIPS3		31
#define	LIPS_JOBSTART_COMMANDLEVEL_LIPS4		41
#define	LIPS_JOBSTART_COMMANDLEVEL_LIPS4C		41
#define	LIPS_JOBSTART_COMMANDLEVEL_CAPSL3		30
#define	LIPS_JOBSTART_COMMANDLEVEL_CAPSL4		40

/*** Character code ***/
#define	LIPS_JOBSTART_CHARCODE_PANNEL	0
#define	LIPS_JOBSTART_CHARCODE_JIS		1
#define	LIPS_JOBSTART_CHARCODE_EUC		2
#define	LIPS_JOBSTART_CHARCODE_DEC		3
#define	LIPS_JOBSTART_CHARCODE_JIS_NEC	4
#define	LIPS_JOBSTART_CHARCODE_JIS_AX	5

/*#define	LIPS_JOB_UNIT_SIZE		300*/
#define	LIPS_UNIT_SIZE			720

#define	LIPS_TEXT_TOP_MARGIN	(63.0 / 300.)

#define	LIPS_TOP_MARGIN		((double)0.005 * IPM + LIPS_TEXT_TOP_MARGIN)
#define	LIPS_TOP_MARGIN_VEC	((double)0.005 * IPM)
#define	LIPS_LEFT_MARGIN	((double)0.005 * IPM)

#define SpToHLp(sp)    (int)(((double)(sp) * glfDimension * (double)gdwMagnification / 1000. - LIPS_LEFT_MARGIN + 1.) * (double)LIPS_UNIT_SIZE + .5)
#define SpToVLp(sp)    (int)(((double)(sp) * glfDimension * (double)gdwMagnification / 1000. - LIPS_TOP_MARGIN + 1.) * (double)LIPS_UNIT_SIZE + .5)

#define SpToHVp(sp)    (int)(((double)(sp) * glfDimension * (double)gdwMagnification / 1000. - LIPS_LEFT_MARGIN + 1.) * (double)LIPS_UNIT_SIZE + .5)
#define SpToVVp(sp)    (int)(((double)(sp) * glfDimension * (double)gdwMagnification / 1000. - LIPS_TOP_MARGIN_VEC + 1.) * (double)LIPS_UNIT_SIZE + .5)

#define	JobToText(n)	(int)(((double)(n) / gwDPI * (double)LIPS_UNIT_SIZE) + .5)

#define	TPICtoVec(n)	(int)((double)(n) / 1000. * 720.)

typedef struct tagLipsVecParam {
	char	acParam[7];
	struct tagLipsVecParam* next;
} LipsVecParam;

typedef struct tagKanjiOffset {
	unsigned DWORD	ch;
	double			fOffset;
} KanjiOffset;

BOOL LipsPdlOutChar(WORD ch);
BOOL LipsPdlPrintInfo(void);
BOOL LipsPdlSelectFont(Fonts* pFont);
BOOL LipsPdlPrintStart(void);
BOOL LipsPdlPrintEnd(void);
BOOL LipsPdlChangeFont(unsigned DWORD fnum);
unsigned DWORD LipsPdlPutChar(unsigned DWORD ch);
BOOL LipsPdlPutRule(DWORD a, DWORD b);
BOOL LipsPdlNewPage(WORD wPageNum);
BOOL LipsPdlFeedPage(void);
BOOL LipsPdlSetColor(Color *pColor);
BOOL LipsPdlSpecial(char *pszSpecial);
void LipsPdlError(ErrorCode err);
void LipsPdlError2(ErrorCode err, char *arg1);
BOOL LipsPdlTPICDrawPolygon(DWORD dwPenSize, TPICArray *pta, BOOL bDrawFrame, BOOL bShading, double lfShading, double lfLength, double lfDistance);
BOOL LipsPdlTPICDrawEllipse(DWORD dwPenSize, BOOL bDrawFrame, double lfShading, DWORD cx, DWORD cy, DWORD rx, DWORD ry, double srad, double erad);
BOOL LipsPdlTPICDrawSpline(DWORD dwPenSize, TPICArray *pta, double lfType);

#endif
