/*
 *  dvipr
 *      TeX Printer driver for The Canon LIPS III/IVc, CaPSL III printer
 *
 *      Developed at:
 *          Computer Systems Research Group
 *          Department of Information Science
 *          Fukui University
 *          9-1, Bunkyo-3, Fukui, 910 Japan
 *
 *      Authors:
 *          Yasumasa Koizumi, Atsushi Takagi, Kenichi Sugimoto,
 *          and Minato Kawaguti
 *
 *  The program may be copied and redistributed freely in the spirit of
 *  Free Software Foundation's "Copy Left", provided this title page is
 *  included in all of the source files involved.
 *
 *  Reports on porting to new platforms, bug fixes, etc., are highly
 *  appreciated. e-mail of this kind of communication shhould be addressedd
 *  to:
 *      dvipr@i1nws1.fukui-u.ac.jp
 */

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <assert.h>
#include <unistd.h>
#include "file.h"
#include "pk.h"
#include "error.h"
#include "dvi.h"
#include "funcdef.h"
#ifdef DVIPR_USE_KPATHSEA
#include <kpathsea/tex-glyph.h>
#endif

#undef	PK_TEST

static BOOL bBW;

static BOOL ReadGryph(FILE *fp, int cmd, Fonts *pFont);
static FILE *OpenPK(Fonts* pFont);
static void ReadGryphBody(FILE *fp, int dyn_f, Gryph *pGryph);
static void SetPixel(Gryph* pGryph, unsigned DWORD x, unsigned DWORD y);
static void ReversePixel(void);
static void LineCopy(Gryph *pGryph, unsigned DWORD y, unsigned DWORD rep);
static int ReadValue(FILE *fp, int dyn_f, unsigned DWORD *val);
static void ReadGryphParam(FILE *fp, Gryph* pGryph, CharWidth* pCharWidth, int cmd);
static BOOL ReadPreAmble(FILE *fp, Fonts *pFont);
#ifdef PK_TEST
static void DisplayGryph(Gryph *pGryph);
#endif

#ifdef PK_TEST
int main(void)
{
	Fonts font;

	strcpy(font.acFilePath, "./cmr10.300pk");
/*	strcpy(font.acFilePath, "/usr/local/lib/texmf/fonts/pkpool//cmbx10.300pk");*/
	ReadPK(&font);
	return 0;
}
#endif

BOOL ReadPK(Fonts *pFont)
{
	unsigned	cmd;
	BOOL		bExitFlag;
	FILE		*fp;

	if ((fp = OpenPK(pFont)) == NULL) {
		Error2(ERR_PK_NOTFOUND, pFont->acName);
		return FALSE;
	}
	if (!ReadPreAmble(fp, pFont)) {
		Error2(ERR_PK_CANNOT_READ_PREAMBLE, pFont->acName);
		fclose(fp);
		return FALSE;
	}
	pFont->Type = FONT_PK;
	pFont->pCharWidth = NULL;
	pFont->pGryph = NULL;
	bExitFlag = FALSE;
	while (!bExitFlag) {
		cmd = FpRead1(fp);
		if (PK_XXX1 <= cmd && cmd <= PK_XXX4) {
			FpSeek(fp, FpReadN(fp, cmd - PK_XXX1 + 1), SEEK_CUR);
			continue;
		}
		switch (cmd) {
			case PK_POST:
				bExitFlag = TRUE;
				break;
			case PK_PRE:
			case PK_NO_OP:
				break;
			case PK_YYY:
				FpSeek(fp, 4, SEEK_CUR);
				break;
			default:
				ReadGryph(fp, cmd, pFont);
				break;
		}
	}
	fclose(fp);
	return TRUE;
}

static FILE *OpenPK(Fonts* pFont)
{
#ifdef DVIPR_USE_KPATHSEA
	char					*pszFileName;
	kpse_glyph_file_type	kpGlrphFileType;

	if ((pszFileName = kpse_find_pk(pFont->acName, pFont->wDPI, &kpGlrphFileType)) == NULL)
		return NULL;
/*fprintf(stderr, "pkname=%s\n", pszFileName);*/
	return fopen(pszFileName, "rb");
#else
	int i;
	char	achPKFileName[FILENAME_MAX];
	FILE* fp;

	for (i = -1; i <= 1; i++) {
		sprintf(achPKFileName, "%s/%s.%dpk", TEXFONTS, pFont->acName, pFont->wDPI + i);
		if ((fp = fopen(achPKFileName, "rb")) != NULL)
			return fp;
	}
	return NULL;
#endif
}

static BOOL ReadGryph(FILE *fp, int cmd, Fonts *pFont)
{
	Gryph		*pNewGryph;
	CharWidth	*pNewCharWidth;

	if ((pNewGryph = CreateGryph(pFont)) == NULL || (pNewCharWidth = CreateCharWidth(pFont)) == NULL)
		return FALSE;
	ReadGryphParam(fp, pNewGryph, pNewCharWidth, cmd);
	if ((pNewGryph->pBits = (unsigned char*)malloc(round8(pNewGryph->dwWidth) * pNewGryph->dwHeight)) == NULL) {
		Error(ERR_NOMEMORY);
		return FALSE;
	}
	memset(pNewGryph->pBits, 0, round8(pNewGryph->dwWidth) * pNewGryph->dwHeight);
	bBW = !!(cmd & FLAG_BW);
	ReadGryphBody(fp, (cmd & FLAG_DYNF) >> 4, pNewGryph);
	FpResetNibble();
#ifdef PK_TEST
	DisplayGryph(pNewGryph);
#endif
	return TRUE;
}

static void ReadGryphBody(FILE *fp, int dyn_f, Gryph *pGryph)
{
	unsigned char	ch;
	int		i;
	unsigned DWORD	x, y, val, rep;

	x = y = rep = 0;
	if (dyn_f == 14) {
		bBW = TRUE;
		while (TRUE) {
			ch = FpRead1(fp);
			for (i = 8; i; i--) {
				if (ch & 0x80)
					SetPixel(pGryph, x, y);
				ch <<= 1;
				if (++x >= pGryph->dwWidth) {
					x = 0;
					y++;
					if (y >= pGryph->dwHeight)
						return;
				}
			}
			
		}
	} else
		while (TRUE)
			switch (ReadValue(fp, dyn_f, &val)) {
				case 0:
					for (; val; val--) {
						if (bBW)
							SetPixel(pGryph, x, y);
						if (++x >= pGryph->dwWidth) {
							x = 0;
							y++;
							if (rep) {
								LineCopy(pGryph, y, rep);
								y += rep;
								rep = 0;
							}
							if (y >= pGryph->dwHeight)
								return;
						}
					}
					ReversePixel();
					break;
				case 1:
					rep = 1;
					break;
				case 2:
					ReadValue(fp, dyn_f, &val);
					rep = val;
					break;
			}
}

static void SetPixel(Gryph* pGryph, unsigned DWORD x, unsigned DWORD y)
{
	unsigned char *ptr;

	ptr = pGryph->pBits + round8(pGryph->dwWidth) * y + (x >> 3);
	*ptr |= (0x80 >> (x % 8));
}

static void ReversePixel(void)
{
	bBW = !bBW;
}

static void LineCopy(Gryph *pGryph, unsigned DWORD y, unsigned DWORD rep)
{
	unsigned char *ptr;

	ptr = pGryph->pBits + round8(pGryph->dwWidth) * (y - 1);
	while (rep--) {
		memcpy(ptr +  round8(pGryph->dwWidth), ptr, round8(pGryph->dwWidth));
		ptr += round8(pGryph->dwWidth);
	}
}

static int ReadValue(FILE *fp, int dyn_f, unsigned DWORD *val)
{
	int		ct;
	unsigned char	cDat;

	cDat = FpReadNibble(fp);
	if (!cDat) {
		ct = 1;
		while (!(cDat = FpReadNibble(fp)))
			ct ++;
		*val = cDat;
		while (ct--) {
			*val <<= 4;
			*val += (cDat = FpReadNibble(fp));
		}
		*val -= 15;
		*val += ((13 - dyn_f) << 4) + dyn_f;
	} else if (1 <= cDat && cDat <= dyn_f)
		*val = cDat;
	else if (dyn_f < cDat && cDat <= 13)
		*val = ((cDat - dyn_f - 1) << 4) + (cDat = FpReadNibble(fp)) + dyn_f + 1;
	else if (cDat == 14)
		return 2;
	else if (cDat == 15)
		return 1;
	return 0;
}

static void ReadGryphParam(FILE *fp, Gryph* pGryph, CharWidth* pCharWidth, int cmd)
{
	if (cmd & FLAG_ESS) {
		if ((cmd & FLAG_SIZE) == FLAG_SIZE) {
			/* This is normal pattern */
			FpRead4(fp);
			pCharWidth->dwCharCode = pGryph->dwCharCode = FpRead4(fp);
			pCharWidth->dwWidth = FpRead4(fp);
			pGryph->dwMoveX = FpRead4(fp); pGryph->dwMoveY = FpRead4(fp);
			pGryph->dwWidth = FpRead4(fp); pGryph->dwHeight = FpRead4(fp);
			pGryph->dwHorzOffset = FpRead4(fp); pGryph->dwVertOffset = FpRead4(fp);
		} else {
			/* This is extra short pattern */
			FpRead2(fp);
			pCharWidth->dwCharCode = pGryph->dwCharCode = FpRead1(fp);
			pCharWidth->dwWidth = FpRead3(fp);
			pGryph->dwMoveX = FpRead2(fp); pGryph->dwMoveY = 0;
			pGryph->dwWidth = FpRead2(fp); pGryph->dwHeight = FpRead2(fp);
			pGryph->dwHorzOffset = FpRead2(fp); pGryph->dwVertOffset = FpRead2(fp);
		}
	} else {
		/* This is extra short pattern */
		FpRead1(fp);
		pCharWidth->dwCharCode = pGryph->dwCharCode = FpRead1(fp);
		pCharWidth->dwWidth = FpRead3(fp);
		pGryph->dwMoveX = FpRead1(fp); pGryph->dwMoveY = 0;
		pGryph->dwWidth = FpRead1(fp); pGryph->dwHeight = FpRead1(fp);
		pGryph->dwHorzOffset = (signed char)FpRead1(fp); pGryph->dwVertOffset = (signed char)FpRead1(fp);
	}
}

static BOOL ReadPreAmble(FILE *fp, Fonts *pFont)
{
	unsigned DWORD	ds;
	
	if (FpRead1(fp) != PK_PRE) {
		Error(ERR_PK_INVALID_FORMAT);
		return FALSE;
	}
	if (FpRead1(fp) != 89) {
		Error(ERR_PK_INVALID_FORMAT);
		return FALSE;
	}
	FpSeek(fp, FpRead1(fp), SEEK_CUR);
	ds = FpRead4(fp) / Pow(2., 4);
	/* I use follwing parameter to caliculate magnification, so must be modified here. */
	pFont->dwScaled = (unsigned DWORD)((double)pFont->dwScaled * ((double)ds / (double)pFont->dwDesignSize));
	pFont->dwDesignSize = ds;
	pFont->dwCheckSum = FpRead4(fp);
	FpRead4(fp);		/* pass hppp */
	FpRead4(fp);		/* pass vppp */
	return TRUE;
}
