/*
 *  dvipr
 *      TeX Printer driver for The Canon LIPS III/IVc, CaPSL III printer
 *
 *      Developed at:
 *          Computer Systems Research Group
 *          Department of Information Science
 *          Fukui University
 *          9-1, Bunkyo-3, Fukui, 910 Japan
 *
 *      Authors:
 *          Yasumasa Koizumi, Atsushi Takagi, Kenichi Sugimoto,
 *          and Minato Kawaguti
 *
 *  The program may be copied and redistributed freely in the spirit of
 *  Free Software Foundation's "Copy Left", provided this title page is
 *  included in all of the source files involved.
 *
 *  Reports on porting to new platforms, bug fixes, etc., are highly
 *  appreciated. e-mail of this kind of communication shhould be addressedd
 *  to:
 *      dvipr@i1nws1.fukui-u.ac.jp
 */

#include <stdlib.h>
#include <memory.h>
#include <unistd.h>
#include "file.h"
#include "font.h"
#include "error.h"
#include "dvipdl.h"
#include "dvi.h"
#include "newtypes.h"
#include "funcdef.h"
#ifdef DVIPR_USE_KPATHSEA
#include <kpathsea/tex-file.h>
#endif

#undef	TFM_TEST

static FILE *OpenTFM(Fonts* pFont);
static BOOL ReadTFMBody(FILE *fp, Fonts *pFont);
static BOOL ReadJTFMBody(FILE *fp, Fonts *pFont, unsigned WORD id);
static BOOL SetParam(unsigned WORD wCode, Fonts* pFont, unsigned DWORD dwWidth);
static void ReadHeader(FILE* fp, Fonts *pFont, unsigned WORD wItems);
static void Read4Param(FILE* fp, unsigned DWORD *ptr, unsigned WORD wItems);

#ifdef TFM_TEST
int main(void)
{
	Fonts font;

	strcpy(font.acFilePath, "./cmr10.tfm");
/*	strcpy(font.acFilePath, "/usr/local/lib/texmf/fonts/pkpool//cmbx10.300pk");*/
	ReadTFM(&font);
	return 0;
}
#endif

BOOL ReadTFM(Fonts *pFont)
{
	unsigned WORD	id;
	BOOL			bRet;
	FILE*			fp;

	if ((fp = OpenTFM(pFont)) == NULL) {
		Error2(ERR_TFM_NOTFOUND, pFont->acName);
		return FALSE;
	}
	id = FpRead2(fp);
	bRet = ((id == 11 || id == 9) ? ReadJTFMBody(fp, pFont, id) : ReadTFMBody(fp, pFont));
	fclose(fp);
	return bRet;
}

static FILE *OpenTFM(Fonts* pFont)
{
#ifdef DVIPR_USE_KPATHSEA
	char *ptr;

	if ((ptr = kpse_find_tfm(pFont->acName)) == NULL)
		return NULL;
/*fprintf(stderr, "tfmname=%s\n", ptr);*/
	return fopen(ptr, "rb");
#else
	char    achTFMFileName[FILENAME_MAX];

	sprintf(achTFMFileName, "%s/%s.tfm", TEXFONTS, pFont->acName);
	return fopen(achTFMFileName, "rb");
#endif
}

static BOOL ReadTFMBody(FILE *fp, Fonts *pFont)
{	
	unsigned WORD	lh, bc, ec, nw, nh, nd, ni, nl, nk, ne, np, ct;
	unsigned DWORD	*pdwWidth, *pdwCharInfo;

	lh = FpRead2(fp);
	bc = FpRead2(fp); ec = FpRead2(fp);
	nw = FpRead2(fp); nh = FpRead2(fp);
	nd = FpRead2(fp); ni = FpRead2(fp);
	nl = FpRead2(fp); nk = FpRead2(fp);
	ne = FpRead2(fp); np = FpRead2(fp);
	ReadHeader(fp, pFont, lh);
	if((pdwWidth = (unsigned DWORD*)malloc(sizeof (DWORD) * nw)) == NULL) {
		Error(ERR_NOMEMORY);
		return FALSE;
	}
	if((pdwCharInfo = (unsigned DWORD*)malloc(sizeof (DWORD) * (ec - bc + 1))) == NULL) {
		Error(ERR_NOMEMORY);
		return FALSE;
	}
	FpSeek(fp, (6 + lh + ec - bc + 1) * 4, SEEK_SET);
	Read4Param(fp, pdwWidth, nw);
	FpSeek(fp, (6 + lh) * 4, SEEK_SET);
	Read4Param(fp, pdwCharInfo, ec - bc + 1);
	for (ct = bc; ct <= ec; ct++)
		SetParam(ct, pFont, pdwWidth[(pdwCharInfo[ct - bc] >> 24) & 0xff]);
	free(pdwCharInfo);
	free(pdwWidth);
	return TRUE;
}

static BOOL ReadJTFMBody(FILE *fp, Fonts *pFont, unsigned WORD id)
{	
	unsigned WORD	nt, lf, lh, bc, ec, nw, nh, nd, ni, nl, nk, ne, np, ct;
	unsigned DWORD	*pdwWidth, *pdwHeight, *pdwCharInfo;
	DWORD			dwCharType;

	nt = FpRead2(fp);
	lf = FpRead2(fp); lh = FpRead2(fp);
	bc = FpRead2(fp); ec = FpRead2(fp);
	nw = FpRead2(fp); nh = FpRead2(fp);
	nd = FpRead2(fp); ni = FpRead2(fp);
	nl = FpRead2(fp); nk = FpRead2(fp);
	ne = FpRead2(fp); np = FpRead2(fp);
	ReadHeader(fp, pFont, lh);
	if((pdwWidth = (unsigned DWORD*)malloc(sizeof (DWORD) * nw)) == NULL) {
		Error(ERR_NOMEMORY);
		return FALSE;
	}
	if((pdwHeight = (unsigned DWORD*)malloc(sizeof (DWORD) * nh)) == NULL) {
		Error(ERR_NOMEMORY);
		return FALSE;
	}
	if((pdwCharInfo = (unsigned DWORD*)malloc(sizeof (DWORD) * (ec - bc + 1))) == NULL) {
		Error(ERR_NOMEMORY);
		return FALSE;
	}
	FpSeek(fp, (7 + lh + nt + ec - bc + 1) * 4, SEEK_SET);
	Read4Param(fp, pdwWidth, nw);
	FpSeek(fp, (7 + lh + nt + ec - bc + 1 + nw) * 4, SEEK_SET);
	Read4Param(fp, pdwHeight, nh);
	FpSeek(fp, (7 + lh + nt) * 4, SEEK_SET);
	Read4Param(fp, pdwCharInfo, ec - bc + 1);
	FpSeek(fp, (7 + lh) * 4, SEEK_SET);
	for (ct = nt; ct; ct--) {
		dwCharType = FpRead4(fp);
		SetParam((dwCharType >> 16) & 0xffff, pFont, pdwWidth[(pdwCharInfo[dwCharType & 0xffff] >> 24) & 0xff]);
	}
	free(pdwCharInfo);
	free(pdwWidth);
	return TRUE;
}

static BOOL SetParam(unsigned WORD wCode, Fonts* pFont, unsigned DWORD dwWidth)
{
	CharWidth*	pNewCharWidth;

	if ((pNewCharWidth = CreateCharWidth(pFont)) == NULL)
		return FALSE;
	pNewCharWidth->dwCharCode = wCode;
	pNewCharWidth->dwWidth = dwWidth;
	return TRUE;
}

static void ReadHeader(FILE* fp, Fonts *pFont, unsigned WORD wItems)
{
	pFont->dwCheckSum = FpRead4(fp);
	pFont->dwDesignSize = FpRead4(fp) / Pow(2., 4);
}

static void Read4Param(FILE* fp, unsigned DWORD *ptr, unsigned WORD wItems)
{
	for (; wItems; wItems--)
		*ptr++ = FpRead4(fp);
}
