/*
 *  dvipr
 *      TeX Printer driver for The Canon LIPS III/IVc, CaPSL III printer
 *
 *      Developed at:
 *          Computer Systems Research Group
 *          Department of Information Science
 *          Fukui University
 *          9-1, Bunkyo-3, Fukui, 910 Japan
 *
 *      Authors:
 *          Yasumasa Koizumi, Atsushi Takagi, Kenichi Sugimoto,
 *          and Minato Kawaguti
 *
 *  The program may be copied and redistributed freely in the spirit of
 *  Free Software Foundation's "Copy Left", provided this title page is
 *  included in all of the source files involved.
 *
 *  Reports on porting to new platforms, bug fixes, etc., are highly
 *  appreciated. e-mail of this kind of communication shhould be addressedd
 *  to:
 *      dvipr@i1nws1.fukui-u.ac.jp
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "pdl.h"
#include "error.h"
#include "funcdef.h"
#include "newtypes.h"
#include "tpic.h"

#define	TOKEN_STR	" \t"
#define	SPLINE_MAX	(256)

static DWORD dwPenSize = 0;
static double lfShading = 0.5;
static TPICArray *pTPICArrayRoot = NULL;

static TPICArray *CreateNewTPICArray(void);
static void DeleteTPICArray(void);

BOOL SpecialTPIC(char *pszSpecial)
{
	TPICArray *pNewTPICArray;
	DWORD cx, cy, rx, ry;
	double s, e;
	char *ptr;

	if (!strncmp(pszSpecial, "pn", 2)) {
		if ((ptr = strtok(&pszSpecial[2], TOKEN_STR)) == NULL)
			return FALSE;
		dwPenSize = atoi(ptr);
		return TRUE;
	} else if (!strncmp(pszSpecial, "pa", 2)) {
		if ((pNewTPICArray = CreateNewTPICArray()) == NULL)
			return FALSE;
		if ((ptr = strtok(&pszSpecial[2], TOKEN_STR)) == NULL)
			return FALSE;
		pNewTPICArray->x = atoi(ptr);
		if ((ptr = strtok(NULL, TOKEN_STR)) == NULL)
			return FALSE;
		pNewTPICArray->y = atoi(ptr);
		return TRUE;
	} else if (!strncmp(pszSpecial, "fp", 2)) {
		gpfnPdlTPICDrawPolygon(dwPenSize, pTPICArrayRoot, TRUE, TRUE, lfShading, 0., 0.);
		lfShading = 0.5;
		DeleteTPICArray();
		return TRUE;
	} else if (!strncmp(pszSpecial, "ip", 2)) {
		gpfnPdlTPICDrawPolygon(dwPenSize, pTPICArrayRoot, FALSE, TRUE, lfShading, 0., 0.);
		lfShading = 0.5;
		DeleteTPICArray();
		return TRUE;
	} else if (!strncmp(pszSpecial, "da", 2)) {
		if ((ptr = strtok(&pszSpecial[2], TOKEN_STR)) == NULL)
			return FALSE;
		gpfnPdlTPICDrawPolygon(dwPenSize, pTPICArrayRoot, TRUE, FALSE, lfShading, atof(ptr), 0.);
		DeleteTPICArray();
		return TRUE;
	} else if (!strncmp(pszSpecial, "dt", 2)) {
		if ((ptr = strtok(&pszSpecial[2], TOKEN_STR)) == NULL)
			return FALSE;
		gpfnPdlTPICDrawPolygon(dwPenSize, pTPICArrayRoot, TRUE, FALSE, lfShading, 0., atof(ptr));
		DeleteTPICArray();
		return TRUE;
	} else if (!strncmp(pszSpecial, "sp", 2)) {
		if ((ptr = strtok(&pszSpecial[2], TOKEN_STR)) == NULL)
			gpfnPdlTPICDrawSpline(dwPenSize, pTPICArrayRoot, 0.0);
		else
			gpfnPdlTPICDrawSpline(dwPenSize, pTPICArrayRoot, atof(ptr));
		DeleteTPICArray();
		return TRUE;
	} else if (!strncmp(pszSpecial, "ar", 2)) {
		if ((ptr = strtok(&pszSpecial[2], TOKEN_STR)) == NULL)
			return FALSE;
		cx = atoi(ptr);
		if ((ptr = strtok(NULL, TOKEN_STR)) == NULL)
			return FALSE;
		cy = atoi(ptr);
		if ((ptr = strtok(NULL, TOKEN_STR)) == NULL)
			return FALSE;
		rx = atoi(ptr);
		if ((ptr = strtok(NULL, TOKEN_STR)) == NULL)
			return FALSE;
		ry = atoi(ptr);
		if ((ptr = strtok(NULL, TOKEN_STR)) == NULL)
			return FALSE;
		s = atof(ptr);
		if ((ptr = strtok(NULL, TOKEN_STR)) == NULL)
			return FALSE;
		e = atof(ptr);
		gpfnPdlTPICDrawEllipse(dwPenSize, TRUE, lfShading, cx, cy, rx, ry, s, e);
		lfShading = 0.5;
		return TRUE;
	} else if (!strncmp(pszSpecial, "ia", 2)) {
		if ((ptr = strtok(&pszSpecial[2], TOKEN_STR)) == NULL)
			return FALSE;
		cx = atoi(ptr);
		if ((ptr = strtok(NULL, TOKEN_STR)) == NULL)
			return FALSE;
		cy = atoi(ptr);
		if ((ptr = strtok(NULL, TOKEN_STR)) == NULL)
			return FALSE;
		rx = atoi(ptr);
		if ((ptr = strtok(NULL, TOKEN_STR)) == NULL)
			return FALSE;
		ry = atoi(ptr);
		if ((ptr = strtok(NULL, TOKEN_STR)) == NULL)
			return FALSE;
		s = atof(ptr);
		if ((ptr = strtok(NULL, TOKEN_STR)) == NULL)
			return FALSE;
		e = atof(ptr);
		gpfnPdlTPICDrawEllipse(dwPenSize, FALSE, lfShading, cx, cy, rx, ry, s, e);
		lfShading = 0.5;
		return TRUE;
	} else if (!strncmp(pszSpecial, "sh", 2)) {
		if ((ptr = strtok(&pszSpecial[2], TOKEN_STR)) == NULL)
			lfShading = 0.5;
		else
			lfShading = atof(ptr);
		return TRUE;
	} else if (!strncmp(pszSpecial, "wh", 2)) {
		lfShading = 0.0;
		return TRUE;
	} else if (!strncmp(pszSpecial, "bk", 2)) {
		lfShading = 1.0;
		return TRUE;
	}
	return FALSE;
}

static TPICArray *CreateNewTPICArray(void)
{
	TPICArray *pNewTPICArray, *pta;

	if ((pNewTPICArray = (TPICArray*)malloc(sizeof (TPICArray))) == NULL) {
		Error(ERR_NOMEMORY);
		return NULL;
	}
	pNewTPICArray->next = NULL;
	if (pTPICArrayRoot == NULL)
		pTPICArrayRoot = pNewTPICArray;
	else {
		for (pta = pTPICArrayRoot; pta->next; pta = pta->next);
		pta->next = pNewTPICArray;
	}
	return pNewTPICArray;
}

static void DeleteTPICArray(void)
{
	TPICArray *pta, *next;

	for (pta = pTPICArrayRoot; pta; pta = next) {
		next = pta->next;
		free(pta);
	}
	pTPICArrayRoot = NULL;
}
